from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from pydelling.managers import PflotranManager, PflotranStudy

from pydelling.managers.callbacks.BaseCallback import BaseCallback
from pathlib import Path


class PflotranRestartCallback(BaseCallback):
    """Callback to restart Pflotran simulations."""
    def __init__(self, manager: PflotranManager, study: PflotranStudy, kind: str = 'post'):
        super().__init__(manager, study, 'pre')

    def run(self):
        """This method should detect the hdf5 file in the previous study and copy it to the current study"""
        if self.study.idx > 0:
            prev_study: PflotranStudy = list(self.manager.studies.values())[self.study.idx - 1]
        else:
            return
        output_files = list(prev_study.output_folder.glob('*.h5'))
        target_file = None
        print(output_files)
        for file in output_files:
            if 'restart' in file.name:
                target_file = file
        if target_file is None:
            raise FileNotFoundError('Restart file not found')
        else:
            self.study.add_input_file(target_file)

    def run_dummy(self):
        """This method is called when the callback is run in dummy mode"""
        pass






class UnitConverter:
    """A class for converting between units of measurement."""
    def convert_time(self, value, initial_unit, final_unit):
        """Converts between time units."""
        value = float(value)
        if initial_unit == 's' and final_unit == 'd':
            return value / 86400
        elif initial_unit == 'd' and final_unit == 's':
            return value * 86400
        elif initial_unit == 's' and final_unit == 'y':
            return value / 31557600
        elif initial_unit == 'y' and final_unit == 's':
            return value * 31557600
        elif initial_unit == 'd' and final_unit == 'y':
            return value / 365
        elif initial_unit == 'y' and final_unit == 'd':
            return value * 365
        elif initial_unit == 'min' and final_unit == 'y':
            return value / 525600
        elif initial_unit == 'y' and final_unit == 'min':
            return value * 525600
        elif initial_unit == 'min' and final_unit == 'd':
            return value / 1440
        elif initial_unit == 'd' and final_unit == 'min':
            return value * 1440
        elif initial_unit == 'min' and final_unit == 's':
            return value / 60
        elif initial_unit == 's' and final_unit == 'min':
            return value * 60
        elif initial_unit == 'min' and final_unit == 'h':
            return value / 60
        elif initial_unit == 'h' and final_unit == 'min':
            return value * 60
        elif initial_unit == 'h' and final_unit == 'd':
            return value / 24
        elif initial_unit == 'd' and final_unit == 'h':
            return value * 24
        elif initial_unit == 'h' and final_unit == 's':
            return value * 3600
        elif initial_unit == 's' and final_unit == 'h':
            return value / 3600
        elif initial_unit == 'h' and final_unit == 'y':
            return value / 8760
        elif initial_unit == 'y' and final_unit == 'h':
            return value * 8760
        elif initial_unit == final_unit:
            return value
        else:
            raise ValueError(f"Invalid unit conversion: {initial_unit} to {final_unit}")








import logging
import pathlib

import dill
import numpy as np
import math
from tqdm import tqdm

import pydelling.preprocessing.mesh_preprocessor.geometry as geometry
import pydelling.preprocessing.mesh_preprocessor.geometry.BaseElement
from pydelling.preprocessing.dfn_preprocessor import DfnPreprocessor
from pydelling.preprocessing.dfn_preprocessor import Fracture, Fault
from pydelling.preprocessing.mesh_preprocessor import MeshPreprocessor
from pydelling.utils.geometry_utils import compute_polygon_area

logger = logging.getLogger(__name__)


class DfnUpscaler:
    def __init__(self, dfn: DfnPreprocessor,
                 mesh: MeshPreprocessor,
                 parallel=False,
                 save_intersections=False,
                 load_faults:str or pathlib.Path=None,
                 loading=False,
                 nearest=None,
                 check_nodes=False,
                 ):
        self.eps = 1E-4
        self.dfn: DfnPreprocessor = dfn
        self.mesh: MeshPreprocessor = mesh
        logger.info('The DFN and mesh objects have been set properly')
        self.all_intersected_points = []
        self.save_intersections = save_intersections
        self.load_faults = load_faults

        self.target_num = 15
        self.cur_num = 0
        self.add_to_class('nearest', nearest, default=None)
        self.add_to_class('check_nodes', check_nodes, default=False)

        if not loading:
            self._intersect_dfn_with_mesh(parallel=parallel)



    def _intersect_dfn_with_mesh(self, parallel=False):
        """Runs the DfnUpscaler"""
        logger.info('Upscaling the DFN to the mesh')
        self.break_for = False
        for fracture in tqdm(self.dfn, desc='Intersecting fractures with mesh', total=len(self.dfn)):
            self.find_intersection_points_between_fracture_and_mesh(fracture)
            # if fracture.local_id == 94:
            #     break
            # if self.break_for:
            #     break

        if not self.load_faults:
            self.find_fault_cells(nearest=self.nearest, check_nodes=self.check_nodes)
        else:
            logger.info(f'Loading fault assignment information from {self.load_faults}')
            with open(self.load_faults, 'rb') as f:
                fault_info = dill.load(f)
                for element in self.mesh.elements:
                    element.associated_faults = fault_info[element.local_id]


        if self.save_intersections:
            import csv
            with open('intersections.csv', 'w') as f:
                writer = csv.writer(f)
                writer.writerow(['x', 'y', 'z', 'value'])
                local_id = 0
                for intersection in self.all_intersected_points:
                    for point in intersection:
                        writer.writerow([point.x, point.y, point.z, local_id])
                    local_id += 1

    def find_intersection_points_between_fracture_and_mesh(self, fracture: Fracture, export_stats=True):
        """Finds the intersection points between a fracture and the mesh"""

        intersection_points = []
        kd_tree_filtered_elements = self.mesh.get_closest_mesh_elements(fracture.centroid, distance=fracture.size)
        counter = 0
        elements_filtered = 0
        for element in kd_tree_filtered_elements:
            element: geometry.BaseElement
            counter += 1
            absolute_distance = np.abs(fracture.distance_to_point(element.centroid))
            characteristic_length = np.power(element.volume, 1 / 3)
            if absolute_distance > 1.25 * characteristic_length:
                elements_filtered += 1
                continue

            intersection_points = element.intersect_with_fracture(fracture)
            if len(intersection_points) > 0:
                a = 1
            if self.save_intersections:
                self.all_intersected_points.append(intersection_points)

            # intersection_area = compute_polygon_area(intersection_points)
            intersection_area = np.abs(compute_polygon_area(intersection_points))
            fracture.intersection_dictionary[element.local_id] = intersection_area
            if len(intersection_points) > 0:
                element.associated_fractures[fracture.local_id] = {
                    'area': intersection_area,
                    'volume': intersection_area * fracture.aperture,
                    'fracture': fracture.local_id,
                }
            n_intersections = len(intersection_points)
            if not n_intersections in self.mesh.find_intersection_stats['intersection_points'].keys():
                self.mesh.find_intersection_stats['intersection_points'][n_intersections] = 0
            self.mesh.find_intersection_stats['intersection_points'][n_intersections] += 1
            self.mesh.find_intersection_stats['total_intersections'] += 1

        self.mesh.is_intersected = True
        return intersection_points

    def find_fault_cells(self, save_fault_cells=True,
                         nearest=None,
                         check_nodes=False,
                         ):
        """Finds the fault cells in the mesh"""
        logger.info('Finding fault cells')
        fault_cells = {}
        for fault in tqdm(self.dfn.faults, desc='Finding distances to faults'):
            fault: Fault
            # Iterate over each triangle individually and find close mesh elements
            triangle_centers = fault.trimesh_mesh.triangles_center
            triangle_areas = fault.trimesh_mesh.area_faces
            characteristic_distance = fault.effective_aperture if fault.effective_aperture is not None else fault.aperture
            logger.info(f'Processing fault {fault.local_id} containing {len(triangle_centers)} triangles')
            close_triangles = []
            for triangle_center in triangle_centers:
                if not nearest:
                    kd_tree_filtered_elements = self.mesh.get_closest_mesh_elements(triangle_center, distance=characteristic_distance)
                    if len(kd_tree_filtered_elements) == 0:
                        continue
                else:
                    nearest: int
                    assert type(nearest) == int, 'nearest must be an integer'
                    kd_tree_filtered_elements = self.mesh.get_closest_n_mesh_elements(triangle_center, n=nearest)
                close_triangles.append(kd_tree_filtered_elements)
            close_triangles = [item for sublist in close_triangles for item in sublist]
            # Filter out duplicates
            temp_dict = {triangle.local_id: triangle for triangle in close_triangles}
            close_triangles = list(temp_dict.values())

            kd_tree_centroids = np.array([elem.centroid for elem in close_triangles])
            # Add element nodes


            logger.info(f'Found {len(kd_tree_centroids)} close elements, computing distances to mesh.')
            distances = fault.distance(kd_tree_centroids)

            for element, distance in zip(close_triangles, distances):
                distance = np.abs(distance)
                if not nearest:
                    if distance < fault.aperture / 2:
                        element.associated_faults[fault.local_id] = {
                            'distance': distance,
                        }
                        fault.associated_elements.append(element)

                else:
                    element.associated_faults[fault.local_id] = {
                        'distance': distance if distance > self.eps else self.eps,
                    }
                    fault.associated_elements.append(element)

            # Check node distances
            if check_nodes:
                logger.info('Checking node distances')
                number_of_nodes = []
                all_nodes = []
                for element in close_triangles:
                    element: pydelling.preprocessing.mesh_preprocessor.geometry.BaseElement
                    nodes = element.coords
                    number_of_nodes.append(len(nodes))
                    all_nodes += [node for node in nodes]
                # all_nodes = np.array(all_nodes).reshape(-1, 3)
                distances = fault.distance(np.array(all_nodes))
                reconstructed_distances = []
                cum_idx = 0
                for n_node in number_of_nodes:
                    reconstructed_distances.append(distances[cum_idx:cum_idx + n_node])
                    cum_idx += n_node

                for element_id, element in enumerate(close_triangles):
                    element: pydelling.preprocessing.mesh_preprocessor.geometry.BaseElement
                    distances = reconstructed_distances[element_id]
                    for distance in distances:
                        if distance < fault.aperture / 2:
                            element.associated_faults[fault.local_id] = {
                                'distance': distance if distance > self.eps else self.eps,
                            }
                            fault.associated_elements.append(element)
                            break

        if save_fault_cells:
            import dill
            for element in self.mesh.elements:
                fault_cells[element.local_id] = element.associated_faults
            with open('fault_cells.pkl', 'wb') as f:
                dill.dump(fault_cells, f)




    def _compute_fracture_volume_in_elements(self):
        # Compute volume of fractures in each element.
        # self.elements.total_fracture_volume = np.zeros([len(elements)])
        for elem in tqdm(self.mesh.elements, desc="Computing fracture volume fractions"):
            for fracture in elem.associated_fractures:
                fracture_dict = elem.associated_fractures[fracture]
                # Attribute of the element: portion of element occupied by fractures.
                elem.total_fracture_volume += fracture_dict['volume']

    def upscale_mesh_porosity(self,
                              matrix_porosity=None,
                              intensity_correction_factor=1.0,
                              existing_fractures_fraction=1.0,
                              truncate_to_min_percentile=5,
                              truncate_to_max_percentile=95,
                              truncate=True,
                              ):
        # Compute upscaled porosity for each element.
        matrix_porosity = 0.0
        self._compute_fracture_volume_in_elements()
        upscaled_porosity = {}
        for elem in tqdm(self.mesh.elements, desc="Upscaling porosity"):
            element_volume = elem.volume
            upscaled_porosity[elem.local_id] = (elem.total_fracture_volume / element_volume) + matrix_porosity * (1 - (elem.total_fracture_volume / element_volume))
            upscaled_porosity[elem.local_id] = np.abs(upscaled_porosity[elem.local_id]) * intensity_correction_factor * (1 / existing_fractures_fraction)

        for fault in self.dfn.faults:
            for element in fault.associated_elements:
                upscaled_porosity[element.local_id] = fault.porosity

        #Post-processing: Truncate values to P5 and P95.
        if truncate:
            resulting_porosity = [upscaled_porosity[local_id] for local_id in upscaled_porosity]
            minimum_porosity = -1.0
            min_percentile = truncate_to_min_percentile - 1
            while minimum_porosity <= 0:
                minimum_porosity = np.percentile(np.array(resulting_porosity)[~np.isnan(resulting_porosity)],
                                                 min_percentile)
                min_percentile = min_percentile + 1
            print("Porosity will be truncated to Percentile = " + str(min_percentile))

            # Truncate to max
            maximum_porosity = np.percentile(np.array(resulting_porosity)[~np.isnan(resulting_porosity)],
                                                truncate_to_max_percentile)  # TODO: check if this is correct (strange results with small datasets)
            maximum_porosity = np.max(np.array(resulting_porosity)[~np.isnan(resulting_porosity)])

            for elem in tqdm(self.mesh.elements, desc="Truncating porosity values to P1 and P99"):

                if math.isnan(upscaled_porosity[elem.local_id]):
                    upscaled_porosity[elem.local_id] = minimum_porosity
                elif upscaled_porosity[elem.local_id] > maximum_porosity:
                    upscaled_porosity[elem.local_id] = maximum_porosity
                elif upscaled_porosity[elem.local_id] <= 0.0:
                    upscaled_porosity[elem.local_id] = minimum_porosity
                else:
                    continue

        vtk_porosity = np.asarray(self.mesh.elements)
        for local_id in upscaled_porosity:
            vtk_porosity[local_id] = upscaled_porosity[local_id]

        refactored_porosity = self.mesh.refactor_array_by_element_type(vtk_porosity)

        self.mesh.cell_data['upscaled_porosity'] = refactored_porosity
        self.upscaled_porosity = upscaled_porosity

        return upscaled_porosity

    def upscale_mesh_storativity(self,
                                 matrix_storativity=None,
                                 truncate_to_min_percentile=5,
                                 truncate_to_max_percentile=95,
                                 truncate=True,
                                 ):

        upscaled_storativity = {}

        for elem in tqdm(self.mesh.elements, desc="Upscaling fractures storativity"):
            upscaled_storativity[elem.local_id] = 0.0
            element_volume = elem.volume
            sum_weights = 0.0
            sum_weigted_storativity = 0.0
            for frac_name in elem.associated_fractures:
                frac_dict = elem.associated_fractures[frac_name]
                frac = frac_dict['fracture']
                frac_volume_in_element = frac_dict['volume'] / element_volume
                sum_weights += frac_volume_in_element
                sum_weigted_storativity += self.dfn[frac].storativity * frac_volume_in_element
                upscaled_storativity[elem.local_id] = sum_weigted_storativity/sum_weights


        for fault in self.dfn.faults:
            for element in fault.associated_elements:
                upscaled_storativity[element.local_id] = fault.storativity


        #Post-processing: Truncate values to P5 and P95.
        if truncate:
            resulting_storativity = [upscaled_storativity[local_id] for local_id in upscaled_storativity]
            minimum_storativity = -1.0
            min_percentile = truncate_to_min_percentile - 1
            while minimum_storativity <= 0:
                minimum_storativity = np.percentile(np.array(resulting_storativity)[~np.isnan(resulting_storativity)], min_percentile)
                min_percentile = min_percentile + 1
                print(min_percentile)

            print("Storativity will be truncated to Percentile = " + str(min_percentile))

            maximum_storativity = np.percentile(np.array(resulting_storativity)[~np.isnan(resulting_storativity)], truncate_to_max_percentile)

            for elem in tqdm(self.mesh.elements, desc="Truncating porosity values to P1 and P99"):
                if math.isnan(upscaled_storativity[elem.local_id]):
                  upscaled_storativity[elem.local_id] = minimum_storativity
                elif upscaled_storativity[elem.local_id] > maximum_storativity:
                    upscaled_storativity[elem.local_id] = maximum_storativity
                elif upscaled_storativity[elem.local_id] <= minimum_storativity:
                    upscaled_storativity[elem.local_id] = minimum_storativity
                else:
                    continue

        vtk_storativity = np.asarray(self.mesh.elements)
        for local_id in upscaled_storativity:
            vtk_storativity[local_id] = upscaled_storativity[local_id]

        self.mesh.cell_data['upscaled_storativity'] = self.mesh.refactor_array_by_element_type(vtk_storativity)
        self.upscaled_storativity = upscaled_storativity

        return upscaled_storativity

    def export_fault_distances(self):
        """Exports the fault distances"""
        distance = {}

        for elem in tqdm(self.mesh.elements, desc="Computing distances"):
            distance[elem.local_id] = 0
            for fault in elem.associated_faults:
                fault_dict = elem.associated_faults[fault]
                distance[elem.local_id] += fault_dict['distance'] if 'distance' in fault_dict else 0

        vtk_distance = np.asarray(self.mesh.elements)
        for local_id in distance:
            vtk_distance[local_id] = distance[local_id]

        self.mesh.cell_data['distance'] =  self.mesh.refactor_array_by_element_type(vtk_distance)
        self.distance = distance

    def export_fracture_property(self, property='area'):
        property_dict = {}
        for elem in tqdm(self.mesh.elements, desc="Computing fracture properties"):
            property_dict[elem.local_id] = 0
            # print(elem.associated_fractures)
            for fracture in elem.associated_fractures:
                fracture_dict = elem.associated_fractures[fracture]
                property_dict[elem.local_id] += fracture_dict[property]

        vtk_property = np.asarray(self.mesh.elements)
        for local_id in property_dict:
            vtk_property[local_id] = property_dict[local_id]

        self.mesh.cell_data[property] = [vtk_property.tolist()]
        self.property_dict = property_dict

    def upscale_mesh_permeability(self,
                                  matrix_permeability=None,
                                  rho=1000,
                                  g=9.8,
                                  mu=8.9e-4,
                                  mode='full_tensor',
                                  truncate_to_min_percentile=5,
                                  truncate_to_max_percentile=95,
                                  truncate=True,
                                  ):

        matrix_permeability = {}

        for elem in tqdm(self.mesh.elements, desc="Creating permeability tensor for dummy anisotropic case"):
            matrix_permeability[elem.local_id] = np.ones([3, 3]) * 0.0

        matrix_permeability_tensor = matrix_permeability

        # Check correct size of matrix_permeability.
        # matrix_permeability_tensor = np.zeros(len(self.elements))
        #
        # if len(matrix_permeability) != len(self.elements):
        #     print("Incorrect size for matrix permeability. Size of variable doesn't match number of elements in the mesh.")
        #     break
        # else:
        #     for elem in tqdm(self.elements, desc="Check size of matrix permeability input"):
        #         if len(matrix_permeability[elem]) == 3:
        #             if np.shape(matrix_permeability[elem]) == (3,3):
        #                 print("Matrix Permeability Tensor (3,3) for Anisotropic case.")
        #                 continue
        #             else:
        #                 print("Matrix Permeability Tensor must be an np.array([3,3]) for Anisotropic case.")
        #         elif len(matrix_permeability[elem]) == 1:
        #             print("Matrix Permeability for Isotropic case.")
        #             matrix_permeability_tensor[elem] = np.zeros([3,3])
        #             matrix_permeability_tensor[elem][0,0] = matrix_permeability[elem]
        #             continue
        #         else:
        #             print("Incorrect Matrix Permeability Tensor. Must be an np.array([3,3]) for use in Anisotropic case or a single float/int for use in Isotropic case.")
        #             continue

        # UPSCALED PERMEABILITY
        fracture_hk = {}
        fault_hk = {}
        upscaled_hk = {}

        # For each fracture, compute permeability tensor,
        # and add it to the elements intersected by the fracture.
        for elem in tqdm(self.mesh.elements, desc="Upscaling fractures permeability"):
            element_volume = elem.volume
            element_porosity = elem.total_fracture_volume / elem.volume
            fracture_hk[elem.local_id] = np.zeros([3, 3])
            upscaled_hk[elem.local_id] = np.zeros([3, 3])
            fault_hk[elem.local_id] = np.zeros([3, 3])

            for frac_name in elem.associated_fractures:
                frac_dict = elem.associated_fractures[frac_name]
                frac = frac_dict['fracture']
                # n1 = math.cos(frac.dip * (math.pi / 180)) * math.sin(frac.dip_dir * (math.pi / 180))
                # n2 = math.cos(frac.dip * (math.pi / 180)) * math.cos(frac.dip_dir * (math.pi / 180))
                # n3 = -1 * math.sin(frac.dip * (math.pi / 180))
                n1 = self.dfn[frac].unit_normal_vector[0]
                n2 = self.dfn[frac].unit_normal_vector[1]
                n3 = self.dfn[frac].unit_normal_vector[2]
                #frac.hk = ((frac.aperture ** 2) * rho * g) / (12 * mu)
                self.dfn[frac].hk = self.dfn[frac].transmissivity / self.dfn[frac].aperture
                #self.dfn[frac].hk = (5.932E-8 * (np.log10(self.dfn[frac].size / 2.0)) ** 2) / self.dfn[frac].aperture

                if mode == 'isotropy':
                    # Add fracture permeability, weighted by the area that the fracture occupies in the element.
                    fracture_hk[elem.local_id][0, 0] += self.dfn[frac].hk * (frac_dict['volume'] / element_volume)  # Kxx

                else:  # 'anisotropy' in 'mode':
                    perm_tensor = np.zeros([3, 3])
                    # for i in range(1, 4):
                    #    for j in range(1, 4):
                    # Compute tensor
                    perm_tensor[0, 0] = self.dfn[frac].hk * ((n2 ** 2) + (n3 ** 2))
                    perm_tensor[0, 1] = self.dfn[frac].hk * (-1) * n1 * n2
                    perm_tensor[0, 2] = self.dfn[frac].hk * (-1) * n1 * n3
                    perm_tensor[1, 1] = self.dfn[frac].hk * ((n3 ** 2) + (n1 ** 2))
                    perm_tensor[1, 2] = self.dfn[frac].hk * (-1) * n2 * n3
                    perm_tensor[2, 2] = self.dfn[frac].hk * ((n1 ** 2) + (n2 ** 2))

                    if 'mode' == 'anisotropy_principals':
                        eigen_perm_tensor = np.diag(np.linalg.eig(perm_tensor)[0])
                        perm_tensor = eigen_perm_tensor

                    frac_volume_in_element = frac_dict['volume'] / element_volume
                    # Add fracture permeability, weighted by the area that the fracture occupies in the element.
                    fracture_hk[elem.local_id][0, 0] += (perm_tensor[0, 0] * frac_volume_in_element)
                    fracture_hk[elem.local_id][0, 1] += (perm_tensor[0, 1] * frac_volume_in_element)
                    fracture_hk[elem.local_id][0, 2] += (perm_tensor[0, 1] * frac_volume_in_element)
                    fracture_hk[elem.local_id][1, 0] += (perm_tensor[0, 1] * frac_volume_in_element)
                    fracture_hk[elem.local_id][1, 1] += (perm_tensor[1, 1] * frac_volume_in_element)
                    fracture_hk[elem.local_id][1, 2] += (perm_tensor[1, 2] * frac_volume_in_element)
                    fracture_hk[elem.local_id][2, 0] += (perm_tensor[0, 1] * frac_volume_in_element)
                    fracture_hk[elem.local_id][2, 1] += (perm_tensor[1, 2] * frac_volume_in_element)
                    fracture_hk[elem.local_id][2, 2] += (perm_tensor[2, 2] * frac_volume_in_element)

            # Sum permeability contribution from fractures and from matrix.

            upscaled_hk[elem.local_id] = fracture_hk[elem.local_id] + matrix_permeability_tensor[elem.local_id] * (
                    1 - element_porosity)

            if len(elem.associated_faults) > 0:
                for fault_name in elem.associated_faults:
                    fault = self.dfn.faults[fault_name]
                    # n1 = math.cos(frac.dip * (math.pi / 180)) * math.sin(frac.dip_dir * (math.pi / 180))
                    # n2 = math.cos(frac.dip * (math.pi / 180)) * math.cos(frac.dip_dir * (math.pi / 180))
                    # n3 = -1 * math.sin(frac.dip * (math.pi / 180))
                    # n1 = fault.unit_normal_vector[0]
                    # n2 = fault.unit_normal_vector[1]
                    # n3 = fault.unit_normal_vector[2]
                    # frac.hk = ((frac.aperture ** 2) * rho * g) / (12 * mu)
                    effective_aperture = fault.effective_aperture if fault.effective_aperture is not None else fault.aperture
                    fault.hk = fault.transmissivity / effective_aperture

                    if mode == 'isotropy':
                        # Add fault permeability.
                        fault_hk[elem.local_id][0, 0] += fault.hk

                #
                # else:  # 'anisotropy' in 'mode':
                #     perm_tensor = np.zeros([3, 3])
                #     # for i in range(1, 4):
                #     #    for j in range(1, 4):
                #     # Compute tensor
                #     perm_tensor[0, 0] = fault.hk * ((n2 ** 2) + (n3 ** 2))
                #     perm_tensor[0, 1] = fault.hk * (-1) * n1 * n2
                #     perm_tensor[0, 2] = fault.hk * (-1) * n1 * n3
                #     perm_tensor[1, 1] = fault.hk * ((n3 ** 2) + (n1 ** 2))
                #     perm_tensor[1, 2] = fault.hk * (-1) * n2 * n3
                #     perm_tensor[2, 2] = fault.hk * ((n1 ** 2) + (n2 ** 2))
                #
                #     if 'mode' == 'anisotropy_principals':
                #         eigen_perm_tensor = np.diag(np.linalg.eig(perm_tensor)[0])
                #         perm_tensor = eigen_perm_tensor
                #
                #     # Add fault permeability. Element porosity equals to 1 when intersected by faults.
                #     fault_hk[elem.local_id][0, 0] += perm_tensor[0, 0]
                #     fault_hk[elem.local_id][0, 1] += perm_tensor[0, 1]
                #     fault_hk[elem.local_id][0, 2] += perm_tensor[0, 1]
                #     fault_hk[elem.local_id][1, 0] += perm_tensor[0, 1]
                #     fault_hk[elem.local_id][1, 1] += perm_tensor[1, 1]
                #     fault_hk[elem.local_id][1, 2] += perm_tensor[1, 2]
                #     fault_hk[elem.local_id][2, 0] += perm_tensor[0, 1]
                #     fault_hk[elem.local_id][2, 1] += perm_tensor[1, 2]
                #     fault_hk[elem.local_id][2, 2] += perm_tensor[2, 2]

            # Sum permeability contribution from faults.
            upscaled_hk[elem.local_id] = upscaled_hk[elem.local_id] + fault_hk[elem.local_id]

        #Post-processing: Truncate values in each direction.
        if truncate:
            for i in range(0,3):
                for j in range(0,3):
                    resulting_hk = [upscaled_hk[local_id][0,0] for local_id in upscaled_hk]
                    minimum_hk = -1.0
                    min_percentile = truncate_to_min_percentile - 1
                    while minimum_hk <= 0:
                        minimum_hk = np.percentile(np.array(resulting_hk)[~np.isnan(resulting_hk)], min_percentile)
                        min_percentile = min_percentile + 1
                        print(min_percentile)

                    print("HK will be truncated to Percentile = " + str(min_percentile))

                    maximum_hk = np.percentile(np.array(resulting_hk)[~np.isnan(resulting_hk)], truncate_to_max_percentile)

                    for elem in tqdm(self.mesh.elements, desc="Truncating permeability values"):
                        if math.isnan(upscaled_hk[elem.local_id][i,j]):
                            upscaled_hk[elem.local_id][i,j] = minimum_hk
                        elif upscaled_hk[elem.local_id][i,j] > maximum_hk:
                            upscaled_hk[elem.local_id][i,j] = maximum_hk
                        elif upscaled_hk[elem.local_id][i,j] <= minimum_hk:
                            upscaled_hk[elem.local_id][i,j] = minimum_hk
                        else:
                            continue

        # Export values to VTK
        vtk_kxx = np.asarray(self.mesh.elements)
        vtk_kyy = np.asarray(self.mesh.elements)
        vtk_kzz = np.asarray(self.mesh.elements)
        vtk_kxy = np.asarray(self.mesh.elements)
        vtk_kxz = np.asarray(self.mesh.elements)
        vtk_kyz = np.asarray(self.mesh.elements)
        for local_id in upscaled_hk:
            vtk_kxx[local_id] = upscaled_hk[local_id][0, 0]
            vtk_kyy[local_id] = upscaled_hk[local_id][1, 1]
            vtk_kzz[local_id] = upscaled_hk[local_id][2, 2]
            vtk_kxy[local_id] = upscaled_hk[local_id][0, 1]
            vtk_kxz[local_id] = upscaled_hk[local_id][0, 2]
            vtk_kyz[local_id] = upscaled_hk[local_id][1, 2]

        self.mesh.cell_data['Kxx'] = self.mesh.refactor_array_by_element_type(vtk_kxx)
        self.mesh.cell_data['Kyy'] = self.mesh.refactor_array_by_element_type(vtk_kyy)
        self.mesh.cell_data['Kzz'] = self.mesh.refactor_array_by_element_type(vtk_kzz)
        self.mesh.cell_data['Kxy'] = self.mesh.refactor_array_by_element_type(vtk_kxy)
        self.mesh.cell_data['Kxz'] = self.mesh.refactor_array_by_element_type(vtk_kxz)
        self.mesh.cell_data['Kyz'] = self.mesh.refactor_array_by_element_type(vtk_kyz)

        self.upscaled_permeability = upscaled_hk

        return upscaled_hk

    def to_vtk(self, filename):
        """Exports the mesh and the upscaled variables to VTK"""
        self.mesh.to_vtk(filename)

    def porosity_to_csv(self, filename='./porosity.csv'):
        """Exports porosity values to csv

        Returns:

        """
        import csv
        logger.info(f"Exporting porosity to {filename}")
        with open(filename, 'w') as csvfile:
            writer = csv.writer(csvfile)
            for element_id in self.upscaled_porosity:
                centroid = self.mesh.elements[element_id].centroid
                porosity = self.upscaled_porosity[element_id]
                writer.writerow([centroid[0], centroid[1], centroid[2], porosity])

    def plot_porosity_histogram(self, filename='upscaled_porosity_histogram.png'):
        """Plots the upscaled porosity histogram."""
        logger.info(f'Plotting upscaled porosity histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_porosity[element_id] for element_id in self.upscaled_porosity], bins=100)
        return fig, ax

    def plot_hkx_histogram(self, filename='upscaled_hkx_histogram.png'):
        """Plots the upscaled hkx histogram."""
        logger.info(f'Plotting upscaled hkx histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_permeability[element_id][0, 0] for element_id in self.upscaled_permeability], bins=100)
        return fig, ax

    def plot_storativity_histogram(self, filename='upscaled_storativity_histogram.png'):
        """Plots the upscaled storativity histogram."""
        logger.info(f'Plotting upscaled storativity histogram to {filename}')
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plt.hist([self.upscaled_storativity[element_id] for element_id in self.upscaled_storativity], bins=100)
        return fig, ax

    def export_intersection_stats(self, filename='intersection_stats.txt'):
        # Export the run_stats dictionary to file
        assert self.mesh.is_intersected, 'The mesh has not been intersected yet.'
        import json
        with open('run_stats.json', 'w') as fp:
            json.dump(self.mesh.find_intersection_stats, fp)

    def save(self, filename='upscaled_model.json'):
        """Save a copy of the class on a serialized pickle object"""
        logger.info(f'Saving a copy of the class to {filename}')
        # Create saving dictionary
        saving_dict = {}
        saving_dict['mesh'] = self.mesh.get_json()
        import jsonpickle
        with open(filename, 'w') as f:
            jsonpickle.encode(self, f)


    def to_json(self, filename='upscaler.json'):
        """Save a copy of the class on a serialized json object"""
        logger.info(f'Saving a copy of the class to {filename}')
        # Create saving dictionary
        saving_dict = {}
        saving_dict['mesh'] = self.mesh.get_json()
        saving_dict['dfn'] = self.dfn.get_json()
        intersected_points = []
        # for point_group in self.all_intersected_points:
        #     intersected_points.append([])
        #     for point in point_group:
        #         intersected_points[-1].append(point.get_json())
        # saving_dict['all_intersected_points'] = intersected_points
        import json
        with open(filename, 'w') as f:
            json.dump(saving_dict, f)

    @classmethod
    def from_json(cls, filename):
        """Load a serialized pickle object"""
        logger.info(f'Loading the upscaling class from {filename}')
        import json
        with open(filename, 'rb') as f:
            load_dict = json.load(f)
            mesh = MeshPreprocessor.from_dict(load_dict['mesh'])
            dfn = DfnPreprocessor.from_dict(load_dict['dfn'])
            loaded_class = cls(mesh=mesh, dfn=dfn, loading=True)
            # loaded_class.all_intersected_points = load_dict['all_intersected_points']
            return loaded_class

    def add_to_class(self, key, value, default=None):
        """Add an attribute to the class"""
        setattr(self, key, value)
        if value != default:
            logger.info(f'Added {key} = {value} to the class')





        





from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class ClipFilter(BaseFilter):
    """
    This class implements the Paraview Integrate Variables filter
    """
    filter_type: str = "Clip"
    counter: int = 0

    def __init__(self, input_filter, name, *args, **kwargs):
        super().__init__(name=name)
        ClipFilter.counter += 1
        self.filter = Clip(Input=input_filter)
        for kwarg in kwargs:
            setattr(self.filter, kwarg, kwargs[kwarg])

    def clip_box(self, box_position, box_length, exact = True):
        """
        This method clips a box like domain into the filter
        Args:
            box_position: bottom left corner coordinates of the box.
            box_length: length vector of the box.
            exact: if true, the box will clip exactly the source filter.
        """
        self.filter.ClipType = 'Box'
        self.filter.ClipType.Position = box_position
        self.filter.ClipType.Length = box_length
        self.filter.Exact = exact



from typing import List

from .GidObject import GidObject
from .Line import Line
from .Point import Point


class Polyline(GidObject):
    """
    This class takes a list of points and creates a collection of lines
    """
    local_id: int = 1

    def __init__(self, points: List[Point], connect=False):
        super().__init__()
        self.lines = []
        self.points = points
        self.connect = connect
        self.set_up()

    def set_up(self):
        for idx, _ in enumerate(self.points[:-1]):
            point_a = self.points[idx]
            point_b = self.points[idx + 1]
            aux_line = Line(point_a, point_b)
            self.lines.append(aux_line)
        if self.connect:
            aux_line = Line(self.points[-1], self.points[0])
            self.lines.append(aux_line)

    def construct(self, *args, **kwargs):
        for point in self.points:
            self.add(point)
        for line in self.lines:
            self.add(line)


"""
Subsurface Fracture Independent Solutions Helper (SubFISH)
"""
import logging

import mpmath as mp
import numpy as np
import scipy.special as scsp

logger = logging.getLogger(__name__)

class SubfishException(Exception):
    pass


def calculate_tang(tang_data):
    logger.info('Calculating Tang solution')
    # Load parameters
    b = tang_data["b"]
    theta = tang_data["poros"]
    tau = tang_data["tort"]
    alpha = tang_data["alpha"]
    D_star = tang_data["Dw"]
    my_lambda = tang_data["lambda"]
    R_prime = tang_data["Rm"]
    R = tang_data["Rf"]
    v = tang_data["v"]
    l = tang_data["l"]

    min_time = float(tang_data["min_time"])
    max_time = float(tang_data["max_time"])
    point_num = int(tang_data["point_num"])
    times = np.geomspace(min_time, max_time, point_num)
    solution = np.zeros(point_num)

    # Derived parameters
    D = alpha * v + D_star
    D_prime = tau * D_star
    nu = v / (2 * D)
    my_beta = np.sqrt(4 * R * D / v ** 2)
    A = b * R / (theta * np.sqrt(R_prime * D_prime))

    A_tilde = l * theta * np.sqrt(R_prime * D_prime) / (v * b)
    B_tilde = R * l / v

    def to_invert(s):
        return mp.exp(nu * l) * mp.exp(
            -nu * l * mp.sqrt(1 + my_beta ** 2 * (mp.sqrt(s + my_lambda) / A + s + my_lambda)))

    def solution_without_disp_insta(t):
        if t <= B_tilde:
            return 0

        exp1 = np.exp(-A_tilde ** 2 / (4 * (t - B_tilde)))
        exp2 = np.exp(-my_lambda * t)
        return A_tilde / (2 * np.sqrt(np.pi) * (t - B_tilde) ** (3 / 2)) * exp1 * exp2

    def solution_without_disp_const(t):
        if t <= B_tilde:
            return 0

        inside_erfc1 = A_tilde / (2 * np.sqrt(t - B_tilde)) - np.sqrt(my_lambda * (t - B_tilde))
        term1 = np.exp(-np.sqrt(my_lambda) * A_tilde) * scsp.erfc(inside_erfc1)

        inside_erfc2 = A_tilde / (2 * np.sqrt(t - B_tilde)) + np.sqrt(my_lambda * (t - B_tilde))
        term2 = np.exp(np.sqrt(my_lambda) * A_tilde) * scsp.erfc(inside_erfc2)

        return 0.5 * np.exp(-B_tilde * my_lambda) * (term1 + term2)

    for step in range(point_num):
        t = times[step]

        if tang_data.get("neglect_long_disp", False):
            # Use analytical expression
            if tang_data["injection_type"] == "instantaneous":
                solution[step] = solution_without_disp_insta(t)
            elif tang_data["injection_type"] == "constant":
                solution[step] = solution_without_disp_const(t)
            else:
                raise SubfishException("Unknown injection_type '" + tang_data["injection_type"] + "'")

        else:
            if tang_data["injection_type"] == "instantaneous":
                solution[step] = mp.invertlaplace(to_invert, t)
            elif tang_data["injection_type"] == "constant":
                solution[step] = mp.invertlaplace(lambda s: to_invert(s) / s, t)
            else:
                raise SubfishException("Unknown injection_type '" + tang_data["injection_type"] + "'")

    return [times, solution]


"""
Base interface for a reader class
"""
import logging

import numpy as np

from pydelling.readers import BaseReader

logger = logging.getLogger(__name__)
import Ofpp
from pydelling.config import config
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


class OpenFoamReader(BaseReader):
    def __init__(self, filename=None):
        self.filename = Path(filename) if filename else Path(config.open_foam_reader.filename)
        logger.info(f"Reading OpenFOAM mesh file from {self.filename}")
        super().__init__(filename=self.filename)


    def open_file(self, filename):
        self.mesh = Ofpp.FoamMesh(self.filename)

    @property
    def cell_centers(self) -> np.array:
        if not config.globals.is_cell_centers_read:
            try:
                self.mesh.read_cell_centres(str(self.filename / "0/C"))
                logger.info(f"Reading cell center locations from {self.filename / '0/C'}")
                config.globals.is_cell_centers_read = True
            except:
                logger.info(f"Reading cell center locations from {self.filename / 'constant/C'}")
                self.mesh.read_cell_centres(str(self.filename / "constant/C"))
                config.globals.is_cell_centers_read = True

        return self.mesh.cell_centres

    @property
    def cell_volumes(self) -> np.array:
        if not config.globals.is_cell_volumes_read:
            try:
                self.mesh.read_cell_volumes(str(self.filename / "0/V"))
                logger.info(f"Reading cell volume locations from {self.filename / '0/V'}")
                config.globals.is_cell_volumes_read = True
            except:
                logger.info(f"Reading cell volume locations from {self.filename / 'constant/V'}")
                self.mesh.read_cell_volumes(str(self.filename / "constant/V"))
                config.globals.is_cell_volumes_read = True

        return self.mesh.cell_volumes

    def get_data(self) -> np.ndarray:
        """
        Outputs the read data
        :return:
        """
        return np.array(0)

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info = {}




from pydelling.paraview_processor.filters import BaseFilter
try:
    from paraview.simple import *
except:
    pass


class TableToPointsFilter(BaseFilter):
    """
    This class implements the Table to Points filter
    """
    filter_type: str = "Table_to_points"
    counter: int = 0

    def __init__(self, filename, name, **params):
        super().__init__(name=name)
        TableToPointsFilter.counter += 1
        self.filter = TableToPoints(Input=str(filename))
        for param in params:
            setattr(self.filter, param, params[param])

    def table2points_columns(self, x_column, y_column, z_column):
        """
        This method sets de xyz columns
        """
        self.filter.Xcolumn = x_column
        self.filter.Ycolumn = y_column
        self.filter.Zcolumn = z_column




class Scalar:
    def __init__(self, value=0.0):
        self.value = value

    def __repr__(self):
        return f"Scalar({self.value})"

    def __str__(self):
        return f"Scalar({self.value})"




"""
Base interface for a reader class
"""
import logging

import numpy as np
import pandas

from pydelling.readers import BaseReader

logger = logging.getLogger(__name__)
from pydelling.config import config
import logging
from pathlib import Path
import h5py
import natsort
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from collections import OrderedDict
from tqdm import tqdm
from pydelling.readers import PflotranProcessingUtils
import colorsys
import seaborn as sns



logger = logging.getLogger(__name__)

class PflotranReader(BaseReader, PflotranProcessingUtils):
    def __init__(self,
                 filename=None,
                 variables=None,
                 ):
        self.filename = Path(filename) if filename else Path(config.pflotran_reader.filename)
        logger.info(f"Reading PFLOTRAN results file from {self.filename}")
        super().__init__(filename=self.filename)
        self.results = {}
        if variables:
            self.variables = variables
        else:
            self.variables = list(self.data[self.time_keys[self.time_values[0]]])
        if 'Material_ID' not in self.variables:
            self.variables.append('Material_ID')

        for time in tqdm(self.time_keys, desc='Reading PFLOTRAN results'):
            data_slice = self.data[self.time_keys[time]]
            data_slice = {key: data_slice[key] for key in self.variables}
            self.results[time] = PflotranResults(time=time, data=data_slice)
        self.variables = list(self.results[self.time_values[0]].variable_keys)

    def open_file(self, filename):
        self.data: h5py.File = h5py.File(self.filename, 'r')
        # Obtain time keys
        self.time_dict_keys = natsort.natsorted(OrderedDict({float(key.split(' ')[2]): key for key in self.data.keys() if 'Time' in key}))
        self.time_keys = OrderedDict({float(key.split(' ')[2]): key for key in self.data.keys() if 'Time' in key})
        self.time_keys = {key: self.time_keys[key] for key in self.time_dict_keys}

    @property
    def time_values(self):
        return natsort.natsorted(self.time_keys.keys())

    @property
    def coordinates(self):
        temp_coordinates = self.data['Coordinates']
        temp_df = {'x[m]': np.array(temp_coordinates['X [m]']),
                    'y[m]': np.array(temp_coordinates['Y [m]']),
                    'z[m]': np.array(temp_coordinates['Z [m]']),
                  }
        return temp_df

    def get_data(self) -> np.ndarray:
        """
        Outputs the read data
        :return:
        """
        return np.array(0)

    def build_info(self):
        """
        Generates a dictionary containing the basic info of the read data
        :return:
        """
        self.info = {}

    def get_observation_point(self, variable, point=(0, 0, 0)):
        """
        Gets the time dependent evolution of a given variable at a given mesh point
        Args:
            variable: variable name to extract the data from
            point: mesh location

        Returns:
            Array containing the temporal evolution of the variable
        """
        temp_array = []
        for time in self.time_values:
            temp_array.append(self.results[time].results[variable][point[0], point[1], point[2]])
        pd_array = pd.DataFrame(np.array(temp_array), columns=[variable])
        return pd_array

    @property
    def mineral_names(self):
        temp_keys = [key.split('_')[0] for key in self.variables if 'VF' in key]
        return temp_keys

    @property
    def species_names(self):
        temp_keys = [key.split('_')[1] for key in self.variables if 'Total' in key]
        return temp_keys


    def get_variable_by_time_index(self, variable: str, time_index: int) -> np.ndarray:
        """
        Returns the results of a given variable at a given time index
        Args:
            variable: variable name
            time_index: time index

        Returns:
            results of the variable at the given time index
        """
        return self.results[self.time_values[time_index]].results[variable]

    def get_variable_by_time(self, variable: str, time: float) -> np.ndarray:
        """
        Returns the results of a given variable at a given time
        Args:
            variable: variable name
            time: time

        Returns:
            results of the variable at the given time
        """
        return self.results[time].results[variable]

    def get_results_by_time(self, time: float) -> dict:
        """
        Returns the results of a given time
        Args:
            time: time

        Returns:
            results of the given time
        """
        return self.results[time].results

    def get_results_by_time_index(self, time_index: int) -> dict:
        """
        Returns the results of a given time index
        Args:
            time_index: time index

        Returns:
            results of the given time index
        """
        return self.results[self.time_values[time_index]].results


    def get_mineral_vf_key(self, mineral):
        """
        Returns the correct key of the mineral volume fraction name
        Args:
            mineral: mineral name

        Returns:
            mineral volume fraction key
        """
        return f"{mineral}_VF [m^3 mnrl_m^3 bulk]"

    def get_mineral_rate_key(self, mineral):
        """
        Returns the correct key of the mineral rate name
        Args:
            mineral: mineral name

        Returns:
            mineral rate key
        """
        return f"{mineral}_Rate [mol_m^3_sec]"

    def get_mineral_si_key(self, mineral):
        """
        Returns the correct key of the mineral si name
        Args:
            mineral: mineral name

        Returns:
            mineral si key
        """
        return f"{mineral}_SI"

    def get_primary_species_key(self, species):
        """
        Returns the correct key of the given species name
        Args:
            species: specie name

        Returns:
            specie key
        """
        return f"Total_{species}"

    def plot_primary_species(self, type='1D', postprocess_dir='./postprocess'):
        """This method plots the primary species of a 1D model"""
        logger.info(f'Plotting [{self.species_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        if type == '1D':
            for species in self.species_names:
                # Generate rate plots
                primary_species_key = self.get_primary_species_key (species)
                plot_df = pd.DataFrame()
                plt.clf()
                for time in config.postprocessing.times:
                    specie_data = self.results[time].results[primary_species_key][:, 0, 0]
                    specie_data_pd = pd.DataFrame (specie_data, columns=[f'{time} years'])
                    plot_df = plot_df.combine_first (specie_data_pd)
                plot_df = plot_df.combine_first (self.x_centroid)
                plot_df = plot_df.set_index ('x[m]')
                plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
                line_plot: plt.Axes = sns.lineplot (data=plot_df)
                line_plot.set_xlabel ('X [m]')
                line_plot.set_ylabel (f'{species}')
                plt.savefig(postprocess_dir / f'{species}.png')

    def plot_vf_variation(self, type='1D', postprocess_dir='./postprocess', ignored_minerals=[]):
        """This method plots the vf variation due to mineral precipitation"""
        logger.info(f'Plotting [{self.mineral_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        postprocess_dir.mkdir(exist_ok=True)
        if type == '1D':
            for mineral in self.mineral_names:
                if mineral in ignored_minerals:
                    continue
                # Generate porosity variation plots
                primary_minerals_key = self.get_mineral_vf_key(mineral)
                plot_df = pd.DataFrame()
                plt.clf()
                # Save initial mineral volumes and porosity
                assert 'Porosity' in self.results[0.0].results, 'Porosity must be within the PFLOTRAN output variables'
                initial_porosity = self.results[0.0].results['Porosity']
                initial_mineral_vf = self.results[0.0].results[primary_minerals_key][:, 0, 0]
                for time in config.postprocessing.times:
                    mineral_vf = self.results[time].results[primary_minerals_key][:, 0, 0]
                    mineral_variation = mineral_vf - initial_mineral_vf
                    mineral_vf_pd = pd.DataFrame (mineral_variation, columns=[f'{time} years'])
                    plot_df = plot_df.combine_first (mineral_vf_pd)
                plot_df = plot_df.combine_first (self.x_centroid)
                plot_df = plot_df.set_index ('x[m]')
                plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
                line_plot: plt.Axes = sns.lineplot (data=plot_df)
                line_plot.set_xlabel ('X [m]')
                line_plot.set_ylabel (f'{mineral} volume fraction variation')
                plt.savefig(postprocess_dir / f'{mineral}.png')

    def plot_total_porosity_variation(self, type='1D', postprocess_dir='./postprocess', ignored_minerals=[]):
        """This method plots the vf variation due to mineral precipitation"""
        logger.info(f'Plotting [{self.mineral_names}] primary species from {self.filename}')
        postprocess_dir = Path(postprocess_dir)
        postprocess_dir.mkdir(exist_ok=True)
        plot_df = pd.DataFrame()

        if type == '1D':
            assert 'Porosity' in self.results[0.0].results, 'Porosity must be within the PFLOTRAN output variables'
            initial_porosity = self.results[0.0].results['Porosity'][:, 0, 0]
            for time in config.postprocessing.times:
                total_porosity_variation = np.zeros_like(initial_porosity)
                for mineral in self.mineral_names:
                    if mineral in ignored_minerals:
                        continue
                    # Generate porosity variation plots
                    primary_minerals_key = self.get_mineral_vf_key(mineral)
                    initial_mineral_vf = self.results[0.0].results[primary_minerals_key][:, 0, 0]
                    mineral_vf = self.results[time].results[primary_minerals_key][:, 0, 0]
                    mineral_variation = mineral_vf - initial_mineral_vf
                    total_porosity_variation += mineral_variation

                total_porosity_variation = total_porosity_variation / initial_porosity * 100
                mineral_vf_pd = pd.DataFrame(total_porosity_variation, columns=[f'{time} years'])
                plot_df = plot_df.combine_first(mineral_vf_pd)
            plot_df = plot_df.combine_first(self.x_centroid)
            plot_df = plot_df.set_index('x[m]')
            plot_df = plot_df.reindex(natsort.natsorted(plot_df.columns), axis=1)
            line_plot: plt.Axes = sns.lineplot (data=plot_df)
            line_plot.set_xlabel ('X [m]')
            line_plot.set_ylabel (f'{mineral} volume fraction variation')
        plt.savefig(postprocess_dir / f'{mineral}.png')

    def plot_1D_slice_of_variable(self, variable,
                                  times=None,
                                  axis='x',
                                  coordinate=0,
                                  postprocess_dir='./postprocess',
                                  color='b',
                                  ) -> plt.Axes:
        """This method plots the variation of a variable along a given axis and coordinate"""
        fig, ax = plt.subplots()
        ax: plt.Axes
        ax_color = ax._get_lines.get_next_color()
        if times is None:
            times = [self.time_values[0]]
        elif times == 'all':
            times = self.time_values
        else:
            times = times

        for time_id, time in enumerate(times):
            data = self.get_variable_by_time(variable, time)
            data_slice = self.get_slice_from_coordinates(data=data,
                                                         axis=axis,
                                                         coordinate=coordinate
                                                         )
            dims = self.get_shape_dimensions(data_slice)
            n_times = len(times)
            # Convert hex to rgb
            next_color = tuple(int(ax_color.lstrip('#')[i:i + 2], 16) / 255 for i in (0, 2, 4))
            original_color = colorsys.rgb_to_hls(*next_color)
            darker_color = colorsys.hls_to_rgb(original_color[0], 0.25 + 0.5 * time_id / n_times, original_color[2])
            data_slice = data_slice.flatten()
            x_data = self.axis_centroids(dims)
            ax.plot(x_data, data_slice, label=f'{time} years', color=darker_color)
            ax.set_xlabel(self.axis_translator[dims])
        ax.set_ylabel(variable)
        ax.grid()
        return ax

    def plot_1D_rigge_variable(self,
                               variable,
                                 times=None,
                                 axis='x',
                                 coordinate=0,
                                 postprocess_dir='./postprocess',
                                 color='b',
                                 ) -> plt.Axes:

        if times is None:
            times = [self.time_values[0]]
        elif times is 'all':
            times = self.time_values
        else:
            times = times

        df = pd.DataFrame()

        for time_id, time in enumerate(times):
            data = self.get_variable_by_time(variable, time)
            data_slice = self.get_slice_from_coordinates(data=data,
                                                         axis=axis,
                                                         coordinate=coordinate
                                                         )
            dims = self.get_shape_dimensions(data_slice)
            n_times = len(times)
            # Convert hex to rgb
            label_name = self.get_shape_dimensions(data_slice)
            data_slice = data_slice.flatten()
            x_data = self.axis_centroids(dims)
            times_var = [time] * len(x_data)
            df = pd.concat([df, pd.DataFrame({'x': x_data, 'y': data_slice, 'times': times_var})])


        a = 2
        sns.set_theme(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})

        # Initialize the FacetGrid object
        pal = sns.cubehelix_palette(len(times), rot=-.25, light=.7)
        g = sns.FacetGrid(df, row="times", hue="times", aspect=15, height=.5, palette=pal)

        g.map(sns.lineplot, "x", 'y',
              clip_on=False,
            alpha=1, linewidth=1.5)
        # Fill the space between the line and the curve
        g.map(plt.fill_between, "x", "y", alpha=.2, clip_on=False)

        # Add a horizontal line to show the maximum value
        max_value = df['y'].max()
        # g.map(plt.axhline, y=max_value, lw=0.5, clip_on=True, color='k')
        def label(x, color, label):
            ax = plt.gca()
            ax.text(-.04, .2, label, fontweight="bold", color=color,
                    ha="left", va="center", transform=ax.transAxes)


        g.map(label, 'x')
        # Set the subplots to overlap
        g.figure.subplots_adjust(hspace=0.5)
        # Change the x axis labels

        # Remove axes details that don't play well with overlap
        g.set_titles("")
        g.set(yticks=[], ylabel="")
        g.despine(bottom=True, left=True)
        g.set_xlabels('x [m]')
        # Set y label in the middle
        g.fig.text(0.01, 0.5, variable, va='center', rotation='vertical')
        # Shrink the plot to fit the legend
        g.fig.subplots_adjust(left=0.1, bottom=0.15)








class PflotranResults:
    """
    This class stores the PFLOTRAN results
    """
    def __init__(self, time, data):
        """
        Args:
            time: timestep
            data: results dataset
        """
        self.time = time
        self.raw_data = data
        self.results = {key: np.array(self.raw_data[key]) for key in self.raw_data}
        self.variable_keys = self.results.keys()

    def __repr__(self):
        return f"Results of time {self.time}"

    @property
    def mineral_names(self):
        temp_keys = [key.split('_')[0] for key in self.variable_keys if 'VF' in key]
        return temp_keys

    @property
    def species_names(self):
        temp_keys = [key.split('_')[1] for key in self.variable_keys if 'Total' in key]
        return temp_keys








from __future__ import annotations

import logging
import os
from collections import OrderedDict
from pathlib import Path
from typing import Dict, List

import h5py
import numpy as np
import pandas as pd

from pydelling.config import config
# from pydelling.readers.iGPReader.geometry import *
from pydelling.preprocessing.mesh_preprocessor.geometry import *
from pydelling.readers.iGPReader.io import BaseReader
from pydelling.readers.iGPReader.utils import get_output_path, RegionOperations
from tqdm import tqdm
from copy import deepcopy

logger = logging.getLogger(__name__)


class iGPReader(BaseReader, RegionOperations):
    """
    This class reads the mesh and region information of an iGP project folder. It also builds an internal representation
    of the mesh, and controls all the pre-processing functions related to the mesh.
    """
    element_dict = {"4": "T", "5": "P", "6": "W", "8": "H"}

    def __init__(self, path, project_name='iGP_project', build_mesh=False, output_folder='./results', write_materials=True):
        self.elements = None
        self.boundaries = {}
        self.element_nodes = None
        logger.info("Initializing iGP Reader module")
        from pydelling.readers.iGPReader.io import AscReader, BoreholeReader, CsvWriter, PflotranExplicitWriter, PflotranImplicitWriter
        self.ExplicitWriter = PflotranExplicitWriter
        self.ImplicitWriter = PflotranImplicitWriter
        self.CsvWriter = CsvWriter
        self.AscReader = AscReader
        self.BoreholeReader = BoreholeReader
        self.path: Path
        if type(path) is not Path:
            self.path = Path(path)
        else:
            self.path = path
        # Read mesh data
        self.is_igp_present = False
        try:
            self.mesh_data = open(self.path / "data.mesh")
            self.region_data = open(self.path / "source.ss")
            self.centroid_data = open(self.path / "centroid.dat")
            self.is_igp_present = True
        except FileNotFoundError:
            self.is_igp_present = False
            logger.info("iGP project files were not found")
        self.build_mesh = build_mesh
        self.project_name = project_name
        self.is_write_materials = write_materials
        self.output_folder = output_folder
        self.mesh_info = {}
        self.region_info = {}
        self.region_dict = {}
        self.material_dict = {}
        self.material_info = {}
        self.centroids = None
        self.is_mesh_built = False
        if not self.build_mesh and self.is_igp_present:
            self.read_mesh_data()
            self.read_region_data()
            self.read_centroid_data()

        if self.build_mesh and self.is_igp_present:
            self.read_mesh_data()
            self.read_region_data()
            self.read_centroid_data()
            self.build_mesh_data()
        logger.info("iGP module initialization complete")
        # Create output folder if it is not created
        if self.output_folder:
            os.makedirs(self.output_folder, exist_ok=True)
        self.mesh_data = None
        self.centroid_data = None
        self.region_data = None

    def read_mesh_data(self):
        """Reads mesh data
        """
        header = self.mesh_data.readline().split()
        self.mesh_info["n_elements"] = int(header[0])
        self.mesh_info["n_nodes"] = int(header[1])
        _elements = []
        for line in range(self.mesh_info["n_elements"]):
            mesh_line = self.mesh_data.readline().split()
            _elements.append([int(element) - 1 for element in mesh_line[1:]])
        # self.elements = np.array(_elements, dtype=np.int32) - 1  # To local ordering
        self.element_nodes = _elements
        assert len(self.element_nodes) == self.mesh_info["n_elements"], "Element number is incorrect"
        # Read node data
        _nodes = []
        for line in range(self.mesh_info["n_nodes"]):
            mesh_line = self.mesh_data.readline().split()
            _nodes.append(mesh_line)
        self.nodes = np.array(_nodes, dtype=np.float32)
        self.nodes_output = _nodes

        assert len(self.nodes) == self.mesh_info["n_nodes"], "Node number is incorrect"
        logger.info(f"Mesh data has been read from {self.path / 'data.mesh'}")

    def initialize_info_dicts(self):
        self._material_names = {}
        for material_id, material in enumerate(self.material_dict):
            self._material_names[material_id] = material
        self.material_info = {}
        for material in self.material_dict:
            self.material_info[material] = {}

    def read_centroid_data(self):
        self.centroids = np.zeros(shape=(len(self.element_nodes), 3))
        for _ in range(self.mesh_info["n_elements"]):
            mesh_line = self.centroid_data.readline().split()
            self.centroids[int(mesh_line[-1]) - 1] = np.array(mesh_line[0:3], dtype=np.float32)
        logger.info(f"Centroid data has been read from {self.path / 'centroid.dat'}")

    def read_region_data(self):
        """
        Reads region and material data
        :return:
        """
        is_reading = True
        header = self.region_data.readline().split()
        self.region_info["n_region"] = int(header[0])
        self.region_info["region_names"] = []
        _temp_array = []
        id_centroids = []
        for i in range(self.region_info["n_region"]):
            line = self.region_data.readline().split()
            self.region_info["region_names"].append(line[0])
        for region_name in self.region_info["region_names"]:
            self.region_dict[region_name] = {"elements": [],
                                             "length": None,
                                             "centroid_id": []}
        # Read region entities
        while is_reading:
            is_element = True
            read_line = self.region_data.readline().split()
            if read_line[0] == "END":
                # self.region_data.readline()  # Skip that line
                is_element = False
                if read_line[1] == "IC":
                    is_reading = False
            if is_element:
                self.region_dict[read_line[-1]]["elements"].append(read_line[2:-1])
                self.region_dict[read_line[-1]]["centroid_id"].append(read_line[1])

        for region_name in self.region_dict:
            self.region_dict[region_name]["elements"] = np.array(self.region_dict[region_name]["elements"],
                                                                 dtype=np.int32) - 1  # Pass to local id
            self.region_dict[region_name]["centroid_id"] = np.array(self.region_dict[region_name]["centroid_id"],
                                                                    dtype=np.int32) - 1  # Pass to local id
            self.region_dict[region_name]["length"] = len(self.region_dict[region_name]["elements"])

        # Read material ids
        self.region_data.readline()  # Skip to start of materials
        is_reading = True  # Activate again the reading of the file
        material_id = 0
        self.material_dict[material_id] = []  # Initialize tuple to add elements
        while is_reading:
            is_material = True  # Assume the line is a material id
            read_line = self.region_data.readline().split()
            if read_line == []:  # The end of the file has been reached
                is_reading = False
                break
            try:
                element_id = int(read_line[0])
                self.material_dict[material_id].append(element_id)
            except ValueError:
                material_name = read_line[0]
                self.material_dict[material_name] = np.array(self.material_dict.pop(material_id))
                material_id += 1
                self.material_dict[material_id] = []  # Initialize new material
        self.material_dict.pop(material_id)  # Take out the last material id (not used)

        logger.info(f"Region data has been read from {self.path / 'source.ss'}")

    # @timer_min
    def implicit_to_explicit(self, dump_mesh_info=True):
        """
        Function that transforms an implicit mesh into an explicit mesh
        :param dump_mesh_info: set it True in order to write the primal mesh into the same unstructured explicit mesh
        """
        if not self.is_mesh_built:
            self.build_mesh_data()
        logger.info("Converting PFLOTRAN implicit mesh to explicit format")
        self.find_connectivities()  # Find the connections of the mesh
        # Write mesh file
        logger.info(f'Writing mesh file to {self.output_folder}')
        exp_mesh_filename = f"{self.project_name}.mesh"
        if self.output_folder is None:
            output_file = open(exp_mesh_filename, "w")
        else:
            output_file = open(os.path.join(self.output_folder, exp_mesh_filename), "w")
        self.ExplicitWriter.write_cells(self, output_file)
        self.ExplicitWriter.write_connections(self, output_file)
        if dump_mesh_info:
            self.ImplicitWriter.write_elements(self, output_file)  # Write elements for mesh visualization
            self.ImplicitWriter.write_nodes(self, output_file)  # Write node_ids for mesh visualization
        output_file.close()
        # Write condition data
        self.ExplicitWriter.write_condition_data(self)
        if self.is_write_materials:
            self.ExplicitWriter.write_materials(self)
        self.write_hdf5_domain()

    def write_explicit_mesh_in_csv(self):
        """
        Writes a .csv file containing the centroids of the mesh elements and each connection's face centers
        """
        logger.info(f"Writing explicit mesh in csv format")
        self.CsvWriter.write_csv_cells(self)
        self.CsvWriter.write_csv_connection(self)

    def write_hdf5_domain(self):
        """
        Writes a hdf5 file containing the domain for post-processing
        """
        if self.output_folder is None:
            hdf5_filename = self.project_name + "-domain.h5"
        else:
            hdf5_filename = os.path.join(self.output_folder, self.project_name + "-domain.h5")
        logger.info(f"Writing mesh to hdf5 ({hdf5_filename})")
        hdf5_file = h5py.File(hdf5_filename, "w")
        self.ExplicitWriter.write_domain_postprocess_hdf5(self, export_file=hdf5_file)
        hdf5_file.close()

    def find_connectivities(self):
        """
        Finds the connectivities of a given implicit mesh. It uses an algorithm that sorts all the faces and finds the
        ones that are in contact to each other, and thus, is able to find connecting cells
        """
        assert self.is_mesh_built, "Mesh is read but not built"
        logger.info("Finding implicit mesh connectivities")
        face_array = []
        for element in self.elements:
            element: BaseElement
            for face in element.faces:
                ordered_face = sorted(element.faces[face].nodes)
                ordered_face.append(element.local_id)
                ordered_face.append(face)
                face_array.append(ordered_face)
        sorted_face_array = sorted(face_array)
        # Connectivity matrix
        conn_dict = {}

        for face_id in range(len(sorted_face_array) - 1):
            face_elements_a = sorted_face_array[face_id][:-2]
            face_elements_b = sorted_face_array[face_id + 1][:-2]
            id_a = sorted_face_array[face_id][-2]
            id_b = sorted_face_array[face_id + 1][-2]
            face_a = sorted_face_array[face_id][-1]
            face_b = sorted_face_array[face_id + 1][-1]
            if face_elements_a == face_elements_b:
                if not id_a in conn_dict:
                    conn_dict[id_a] = {}
                    conn_dict[id_a][id_b] = face_a
                else:
                    conn_dict[id_a][id_b] = face_a
                # if not id_b in conn_dict: # only one way
                #     conn_dict[id_b] = {}
                #     conn_dict[id_b][id_a] = face_b
                # else:
                #     conn_dict[id_b][id_a] = face_b
        conn_dict_ordered = OrderedDict(conn_dict)
        self.connections = sorted(conn_dict_ordered.items())

    def raster_interpolator(self, raster_folder, raster_filenames):
        """Performs layer based raster interpolation
        This function approximates the node_ids that lay on each of the layers with the rasterized data of such layer.
        The current approach is based on nearest neighbours approximation
        :return np.array of interpolated node_ids
        """
        logger.info("Interpolating raster regions to mesh")
        for region in config.raster_refinement.regions:
            self._raster_max_error = 0.0
            logger.info(f"Interpolating region {region}")
            raster_current_region_filename = os.path.join(raster_folder, raster_filenames[region])
            raster_data = self.AscReader(raster_current_region_filename)
            # print(f"Elements in face length is {len(iGP_data.region_dict[region]['elements'])}")
            id_list = np.unique(self.region_dict[region]["elements"].flatten())
            _test = []
            if region in config.raster_refinement.top_regions:
                logger.info(f"An offset of +{config.raster_refinement.top_offset} m will be applied in the z-direction to region {region}")
            if region == config.raster_refinement.second_layer:
                logger.info(f"Region {region} has been selected as the second layer region. Thus, its node_ids will be displaced {config.raster_refinement.first_layer_to_second_layer_difference}m from the top raster file")
            for mesh_id in id_list:
                x_mesh = self.nodes[mesh_id][0]
                y_mesh = self.nodes[mesh_id][1]
                # Raster data
                d_raster = raster_data.info_dict["cellsize"]
                origin_x = raster_data.info_dict["xllcorner"]
                origin_y = raster_data.info_dict["yllcorner"]
                ix = int(np.floor((x_mesh - origin_x) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
                # the floor function
                iy = int(np.floor((y_mesh - origin_y) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
                # the floor function
                iy = int(raster_data.info_dict["ncols"] - iy - 1)
                z_raster = raster_data.data[iy, ix]
                z_offset = config.raster_refinement.top_offset if region in config.raster_refinement.top_regions else 0.0
                interpolation_difference = abs((z_raster + z_offset) - self.nodes[mesh_id][2])
                self._raster_max_error = interpolation_difference if interpolation_difference > self._raster_max_error else self._raster_max_error
                z_second_layer_diff = config.raster_refinement.first_layer_to_second_layer_difference if region == config.raster_refinement.second_layer else 0.0  # Corrects the second layer on a layer based mesh
                self.nodes[mesh_id][2] = z_raster + z_offset + z_second_layer_diff
            logger.info(f"Interpolation of region {region} completed. Max absolute difference {self._raster_max_error:1.2f}m")

        # node_ids = raster_interpolator(self.path, raster_folder, raster_filenames)
        # self.node_ids = node_ids  # Update the mesh with the moved node_ids
        return self.nodes

    def write_ASCII_meshfile(self, filename):
        """
        Writes the mesh in PFLOTRAN ascii file
        :param filename: name of the output filename
        :return:
        """
        # TODO: Set-up default filename
        if self.is_mesh_built:  # We need it to dump it correctly into implicit/explicit format
            if self.output_folder is None:
                file = open(filename, "w")
            else:
                file = open(os.path.join(self.output_folder, filename), "w")
            # Write header
            file.write(f"{self.mesh_info['n_elements']} {self.mesh_info['n_nodes']}\n")
            for element in self.elements:
                file.write(
                    f"{config.globals.element_dict[len(element.nodes)]} {' '.join(map(str, element.nodes + 1))}\n")
            for id, node in enumerate(self.nodes_output):
                file.write(f"{node[0]} {node[1]} {self.nodes[id][2]:5.5f}\n")
                # file.write(f"{node[0]} {node[1]} {node[2]}\n")
            file.write("\n")
            file.close()
        else:  # If mesh is not internally built
            if self.output_folder is None:
                file = open(filename, "w")
            else:
                file = open(os.path.join(self.output_folder, filename), "w")
            file.write(f"{self.mesh_info['n_elements']} {self.mesh_info['n_nodes']}\n")
            for element in self.element_nodes:  # Element data
                file.write(f"{config.globals.element_dict[len(element)]} {' '.join(map(str, element + 1))}\n")
            for id, node in enumerate(self.nodes_output):  # Node coordinates data
                file.write(f"{node[0]} {node[1]} {self.nodes[id][2]:5.5f}\n")
            file.write("\n")
            file.close()
        logger.info(f"Mesh data has been properly exported to PFLOTRAN mesh ASCII format into the {filename} file")

    def write_hdf5_meshfile(self, filename="regions.h5"):
        """
        Export the mesh into implicit hdf5 format.
        :param filename: name of the hdf5 file. Default: "regions.h5"
        """
        hdf5_filename = filename
        if not filename.endswith(".h5"):
            hdf5_filename += ".h5"
        hdf5_file = h5py.File(hdf5_filename, "w")
        self.ImplicitWriter.write_domain_hdf5(self, hdf5_file)
        self.ImplicitWriter.write_regions_hdf5(self, hdf5_file)
        hdf5_file.close()
        logger.info(f"Mesh file has been properly written into '{hdf5_filename}'")

    def set_material_values_from_borehole_data(self):
        logger.info("Setting permeability data from borehole files")
        # assert self.is_mesh_built, "Mesh is not built"
        borehole_reader = self.BoreholeReader(igp_reader=self)
        borehole_reader.run()

    def assign_heterogeneous_materials(self):
        logger.info('Processing heterogeneous material properties')
        for material in config.borehole_processing.heterogeneous_distribution:
            material_properties = {'permeability': [], 'porosity': []}
            for property in config.borehole_processing.heterogeneous_distribution[material]:
                if config.borehole_processing.heterogeneous_distribution[material][property].type == 'constant':
                    material_properties = self._assign_constant_heterogeneous_materials(material=material, property=property, material_properties=material_properties)
                elif config.borehole_processing.heterogeneous_distribution[material][property].type == 'linear':
                    material_properties = self._assign_linear_heterogeneous_materials(material=material, property=property, material_properties=material_properties)
                elif config.borehole_processing.heterogeneous_distribution[material][property].type == 'topography':
                    material_properties = self._assign_topography_heterogeneous_materials(material=material, property=property, material_properties=material_properties)
                elif config.borehole_processing.heterogeneous_distribution[material][property].type == 'step':
                    material_properties = self._assign_step_heterogeneous_materials(material=material, property=property, material_properties=material_properties)
            if material_properties['permeability']:
                self.material_info[material]['heterogeneous_permeability'] = np.array(material_properties['permeability']).astype(np.float)
            if material_properties['porosity']:
                self.material_info[material]['heterogeneous_porosity'] = np.array(material_properties['porosity']).astype(np.float)


    def _assign_constant_heterogeneous_materials(self, material: str, property: str, material_properties: Dict[str, List]) -> Dict[str, List]:
        """
        This method assigns a constant material value to a given material

        Args:
            material: current material to assign properties to
            property: property to be assigned (currently works with porosity and permeabiltiy)
            material_properties: dictionary containing the resulting heterogeneous distributions

        Returns:
            The modified material_properties object
        """
        material_features = config.borehole_processing.heterogeneous_distribution[material][property]
        logger.info(f"Adding constant {property} (value={material_features.value}) to the {material} material")
        for _ in self.material_dict[material] - 1:
            material_properties[property].append(material_features.value)
        return material_properties

    def _assign_step_heterogeneous_materials(self, material: str, property: str, material_properties: Dict[str, List]) -> Dict[str, List]:
        """
        This method assigns a step based material properties to a given material

        Args:
            material: current material to assign properties to
            property: property to be assigned (currently works with porosity and permeabiltiy)
            material_properties: dictionary containing the resulting heterogeneous distributions

        Returns:
            The modified material_properties object
        """
        material_features = config.borehole_processing.heterogeneous_distribution[material][property]
        logger.info(f"Adding step based {property} to the {material} material. Steps: {material_features.z_steps} Values: {material_features.value_list}")
        for centroid in self.centroids[self.material_dict[material] - 1]:
            centroid_bin = np.digitize(centroid[2], material_features.z_steps)
            value = material_features.value_list[centroid_bin - 1]
            material_properties[property].append(value)
        return material_properties

    def _assign_topography_heterogeneous_materials(self, material: str, property: str, material_properties: Dict[str, List]) -> Dict[str, List]:
        """
        This method assigns heterogeneous material properties following that the highest value is set on the topography layer

        Args:
            material: current material to assign properties to
            property: property to be assigned (currently works with porosity and permeabiltiy)
            material_properties: dictionary containing the resulting heterogeneous distributions

        Returns:
            The modified material_properties object
        """
        # Open Topography raster file
        material_features = config.borehole_processing.heterogeneous_distribution[material][property]
        logger.info(
            f"Adding linear {property} value taking into account topography of layer {material_features.topography_region} to material {material} in such a way that f(z_topography) = {material_features.value_top}."
            f"{f' Values below {material_features.z_bot}m are kept constant and equal to {material_features.value_bot}' if material_features.bottom_constant else ''}")
        raster_current_region_filename = Path(config.data_files.raster_file_folder) / config.data_files.raster_filenames[material_features.topography_region]
        raster_data = self.AscReader(raster_current_region_filename)

        for centroid in self.centroids[self.material_dict[material] - 1]:
            x_mesh = centroid[0]  # x-coordinate of centroid
            y_mesh = centroid[1]  # y-coordinate of centroid
            # Raster data
            d_raster = raster_data.info_dict["cellsize"]
            origin_x = raster_data.info_dict["xllcorner"]
            origin_y = raster_data.info_dict["yllcorner"]
            ix = int(np.floor((x_mesh - origin_x) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
            # the floor function
            iy = int(np.floor((y_mesh - origin_y) / (1.001 * d_raster)))  # 1.001 value is used to avoid issues with
            # the floor function
            iy = int(raster_data.info_dict["ncols"] - iy - 1)
            z_offset = config.raster_refinement.top_offset if config.raster_refinement.top_offset else 0.0
            z_raster = raster_data.data[iy, ix] + z_offset

            if material_features.log:
                # z_topography = self.
                value = self.linear_log_distribution(x=centroid[2],
                                                   x_top=z_raster,
                                                   x_bot=material_features.z_bot,
                                                   y_top=material_features.value_top,
                                                   y_bot=material_features.value_bot,
                                                   )
                if material_features.bottom_constant:
                    value = value if centroid[2] > material_features.z_bot else material_features.value_bot
                material_properties[property].append(value)
            else:
                value = self.linear_distribution(x=centroid[2],
                                                   x_top=z_raster,
                                                   x_bot=material_features.z_bot,
                                                   y_top=material_features.value_top,
                                                   y_bot=material_features.value_bot,
                                                   )
                if material_features.bottom_constant:
                    value = value if centroid[2] > material_features.z_bot else material_features.value_bot
                material_properties[property].append(value)
        return material_properties


    def _assign_linear_heterogeneous_materials(self, material: str, property: str, material_properties: Dict[str, List]) -> Dict[str, List]:
        """
        This method assigns heterogeneous properties giving four linear values to a given material

        Args:
            material: current material to assign properties to
            property: property to be assigned (currently works with porosity and permeabiltiy)
            material_properties: dictionary containing the resulting heterogeneous distributions

        Returns:
            The modified material_properties object
        """
        logger.info(f"Adding linear {property} to the {material} material")
        material_features = config.borehole_processing.heterogeneous_distribution[material][property]
        for centroid in self.centroids[self.material_dict[material] - 1]:
            if material_features.log:
                value = self.linear_log_distribution(x=centroid[2],
                                                   x_top=material_features.z_top,
                                                   x_bot=material_features.z_bot,
                                                   y_top=material_features.value_top,
                                                   y_bot=material_features.value_bot,
                                                   )
                if material_features.bottom_constant:
                    value = value if centroid[2] > material_features.z_bot else material_features.value_bot
                material_properties[property].append(value)
            else:
                value = self.linear_distribution(x=centroid[2],
                                                   x_top=material_features.z_top,
                                                   x_bot=material_features.z_bot,
                                                   y_top=material_features.value_top,
                                                   y_bot=material_features.value_bot,
                                                   )
                if material_features.bottom_constant:
                    value = value if centroid[2] > material_features.z_bot else material_features.value_bot
                material_properties[property].append(value)
        return material_properties

    @staticmethod
    def linear_distribution(x, y_bot, y_top, x_bot, x_top):
        return (y_top - y_bot) / (x_top - x_bot) * x + (y_bot * x_top - y_top * x_bot) / (x_top - x_bot)

    @staticmethod
    def linear_log_distribution(x, x_top, x_bot, y_top, y_bot):
        y_top = np.log10(y_top)
        y_bot = np.log10(y_bot)
        return np.power(10.0, (y_top - y_bot) / (x_top - x_bot) * x + (y_bot * x_top - y_top * x_bot) / (x_top - x_bot))

    def export_materials_in_hdf5(self):
        """
        This method uses the information stored in material info to produce an hdf5 file containing heterogeneous data for each material
        """
        idx_set = np.array([])
        perm_dataset = np.array([])
        porosity_dataset = np.array([])

        for material in self.material_info:
            idx_set = np.concatenate([idx_set, self.material_dict[material]]).astype(np.int)
            if 'heterogeneous_permeability' in self.material_info[material]:
                perm_dataset = np.concatenate([perm_dataset, self.material_info[material]['heterogeneous_permeability']])
            if 'heterogeneous_porosity' in self.material_info[material]:
                porosity_dataset = np.concatenate([porosity_dataset, self.material_info[material]['heterogeneous_porosity']])
        # idx_set = idx_set
        if perm_dataset.size != 0:
            assert perm_dataset.min != 0, "Minimum value of index cannot be 0, change to PFLOTRAN numbering [1, N]"
            self._export_hdf5_dataset(index_dataset=idx_set,
                                      dataset=perm_dataset,
                                      name='permeability')
        if porosity_dataset.size != 0:
            assert porosity_dataset.min != 0, "Minimum value of index cannot be 0, change to PFLOTRAN numbering [1, N]"
            self._export_hdf5_dataset(index_dataset=idx_set,
                                      dataset=porosity_dataset,
                                      name='porosity')

    @staticmethod
    def _export_hdf5_dataset(index_dataset: np.array, dataset: np.array, name):
        hdf5_filename = get_output_path() / f"{name}.h5"
        logger.info(f"Writing {name} dataset to hdf5 ({hdf5_filename})")
        with h5py.File(hdf5_filename, "w") as hdf5_file:
            hdf5_file.create_dataset('Cell Ids', data=index_dataset)
            hdf5_file.create_dataset(name, data=dataset)

    def write_csv(self, filename):
        """
        Writes the mesh centroids in .csv format
        :param filename:
        :return:
        """
        # TODO: set-up default filename
        if self.output_folder is None:
            file = open(filename, "w")
        else:
            file = open(os.path.join(self.output_folder, filename), "w")
        # Write header
        for id, node in enumerate(self.nodes_output):
            file.write(f"{node[0]},{node[1]},{self.nodes[id][2]:5.5f}\n")
        file.close()
        logger.info(f"Mesh data has been properly exported to csv format into the '{filename}' file")

    @staticmethod
    def chunks(l, n):
        return [l[i:i + n] for i in range(0, len(l), n)]

    def build_mesh_data(self, processes=1,
                        generate_cells=True,
                        generate_boundaries=True,
                        ):
        """
        Creates an internal representation of the mesh based on a given unstructured implicit grid.
        :return:
        """
        if config.general.constant_centroids:
            logger.info('The location of the centroids will not be changed after refining the boundaries')
        if processes == 1:
            logger.info("Building implicit mesh structure")
            temp = []
            amount_read = 0.0
            if generate_cells:
                for id_local, element in tqdm(enumerate(self.element_nodes), total=len(self.element_nodes), desc='Building mesh'):
                    n_type = len(element)
                    if n_type == 4:  # This is a Wedge object
                        temp.append(TetrahedraElement(node_ids=element,
                                                      node_coords=self.nodes[element],
                                                      local_id=id_local,
                                                      centroid_coords=self.centroids[id_local] if config.general.constant_centroids else None,
                                                      # centroid_coords=self.centroids[id_local]
                                                      ))
                    if n_type == 6:  # This is a Wedge object
                        temp.append(WedgeElement(node_ids=element,
                                                 node_coords=self.nodes[element],
                                                 local_id=id_local,
                                                 centroid_coords=self.centroids[id_local] if config.general.constant_centroids else None,
                                                 # centroid_coords=self.centroids[id_local]
                                                 ))
                    if n_type == 8:  # This is a Hexahedra object
                        temp.append(HexahedraElement(node_ids=element,
                                                     node_coords=self.nodes[element],
                                                     local_id=id_local,
                                                     centroid_coords=self.centroids[id_local] if config.general.constant_centroids else None,
                                                     # centroid_coords=self.centroids[id_local]
                                                     ))
                self.elements = temp
            # Generate boundary information
            if generate_boundaries:
                for boundary in tqdm(self.region_dict, desc='Generating boundary mesh'):
                    temp_boundary = []
                    for element in self.region_dict[boundary]['elements']:
                        n_type = len(element)
                        if n_type == 4:
                            temp_boundary.append(QuadrilateralFace(
                                node_ids=element,
                                node_coords=self.nodes[element],
                            )
                            )
                        if n_type == 3:
                            temp_boundary.append(TriangleFace(
                                node_ids=element,
                                node_coords=self.nodes[element],
                            )
                            )
                    self.boundaries[boundary] = temp_boundary
            self.is_mesh_built = True
        else:
            logger.info(f"Building internal mesh using multiprocessing with {processes} processes")
            from multiprocessing import Process, Manager
            import multiprocessing as mp
            with Manager() as manager:
                shared_list = manager.list()
                processes_list = []
                total_number_of_elements = len(self.element_nodes)
                number_of_processes = processes if processes is not None else mp.cpu_count() - 1
                chunk_size = int(total_number_of_elements / number_of_processes)
                chunks = self.chunks(self.element_nodes, chunk_size)

                for i, chunk in enumerate(chunks):
                    p = Process(target=parallel_build_mesh_data, args=(chunk, self.nodes, shared_list, i, chunk_size, self.centroids))
                    processes_list.append(p)
                for id, process in enumerate(processes_list):
                    process.start()
                    logger.info(f"Process {id} has been started")
                for id, process in enumerate(processes_list):
                    process.join()
                    logger.info(f"Process {id} has finished")
                self.elements = list(shared_list)
            # Order self.elements list following the local_id
            self.elements = sorted(self.elements, key=lambda x: x.local_id)

            self.is_mesh_built = True

    def set_material_z_ranges(self):
        logger.debug("Calculating z-ranges of the materials")
        self.material_ranges = {}
        for material in self.material_dict:
            _material_centroids: np.array = self.centroids[self.material_dict[material] - 1]
            _min_z = np.min(_material_centroids[:, 2])
            _max_z = np.max(_material_centroids[:, 2])
            self.material_info[material] = {"min_z": _min_z, "max_z": _max_z}

    def write_input_raster_files(self, downsample_factor=None, file_format="asc"):
        logger.info(f"Exporting raster files read from {config.data_files.raster_file_folder}")
        for region in config.raster_refinement.regions:
            raster_current_region_filename = os.path.join(config.data_files.raster_file_folder, config.data_files.raster_filenames[region])
            raster_data = self.AscReader(raster_current_region_filename)
            if downsample_factor:
                raster_data.downsample_data(slice_factor=downsample_factor)
            if file_format == 'asc':
                raster_data.dump_to_asc(output_file=get_output_path() / f"{region}.asc")
            elif file_format == "csv":
                raster_data.dump_to_csv(output_file=get_output_path() / f"{region}.csv")

    def export_recharge_file(self, csv_export=False):
        if config.data_files.recharge_file:
            logger.info(f"Processing recharge file located at {config.data_files.recharge_file}")
            recharge_data = self.AscReader(config.data_files.recharge_file)
        else:
            logger.error("A recharge file has to be submitted in the configuration file")
            raise FileNotFoundError("Recharge file was not submitted in the configuration file")
        # Transform recharge data to m/s
        recharge_data.change_data(recharge_data.data / (1000.0 * 365 * 24 * 3600))
        recharge_data.change_data(np.flip(np.transpose(recharge_data.data), axis=1))
        recharge_data.export_to_gridded_dataset(filename="recharge.h5")
        if csv_export:
            recharge_data.dump_to_csv(output_file="recharge.csv")

    def export_region_data(self, regions: List = None):
        regions = regions if regions else config.extract_data.regions
        for region in regions:
            id_list = np.unique(self.region_dict[region]["elements"].flatten())
            node_coordinates_df: pd.DataFrame = pd.DataFrame(self.nodes[id_list], columns=["x", "y", "z"])
            node_coordinates_df.to_csv(get_output_path() / f"{region}_nodes.csv", index=False)

    def get_mesh(self):
        return self.centroids

    @property
    def materials(self) -> pd.DataFrame:
        """
        This property returns a pandas dataframe containing the materials of the iGP model
        Returns:
            A pandas dataframe object containing the materials of the iGP project
        """
        return self.material_dict

    @property
    def n_mesh_elements(self) -> int:
        """
        Gives the total number of mesh elements
        Returns:
            The total number of mesh elements
        """
        return self.mesh_info["n_elements"]

    @property
    def n_mesh_nodes(self) -> int:
        """
        Gives the total number of mesh node_ids
        Returns:
            The total number of mesh node_ids
        """
        return self.mesh_info["n_nodes"]

    def print_material_info(self):
        """
        This method prints the material info in a neat way
        """
        for material in self.material_info:
            step = 0
            print(f"{step * '  '}{material}:")
            for property in self.material_info[material]:
                step = 1
                print(f"{step * '  '}{property} = {self.material_info[material][property]}")

    def get_region_centroids(self, region_name):
        return self.centroids[self.region_dict[region_name]['centroid_id'] - 1]

    def get_region_nodes(self, region_name):
        cur_array = self.region_dict[region_name]['elements']
        cur_array = cur_array.flatten()
        cur_array = np.unique(cur_array)
        return self.nodes[cur_array]

    def get_boundary_faces(self, region_name) -> List[BaseFace]:
        assert self.is_mesh_built, "Mesh has to be built before calling this method"
        return self.boundaries[region_name]

    def get_material_elements(self, material_name) -> List[BaseElement]:
        assert self.is_mesh_built, "Mesh has to be built before calling this method"
        return self.material_dict[material_name]

    def get_material_centroids(self, material_name):
        return self.centroids[self.material_dict[material_name]]

    @property
    def min_x(self):
        '''Returns the minimum x coordinate of the centroids of the mesh'''
        return min(self.centroids[:, 0])

    @property
    def max_x(self):
        '''Returns the maximum x coordinate of the centroids of the mesh'''
        return max(self.centroids[:, 0])

    @property
    def min_y(self):
        '''Returns the minimum y coordinate of the centroids of the mesh'''
        return min(self.centroids[:, 1])

    @property
    def max_y(self):
        '''Returns the maximum y coordinate of the centroids of the mesh'''
        return max(self.centroids[:, 1])

    @property
    def min_z(self):
        '''Returns the minimum z coordinate of the centroids of the mesh'''
        return min(self.centroids[:, 2])

    @property
    def max_z(self):
        '''Returns the maximum z coordinate of the centroids of the mesh'''
        return max(self.centroids[:, 2])

    @property
    def coords_min_x(self):
        '''Returns the minimum x coordinate of the nodes of the mesh'''
        return min(self.nodes[:, 0])

    @property
    def coords_max_x(self):
        '''Returns the maximum x coordinate of the nodes of the mesh'''
        return max(self.nodes[:, 0])

    @property
    def coords_min_y(self):
        '''Returns the minimum y coordinate of the nodes of the mesh'''
        return min(self.nodes[:, 1])

    @property
    def coords_max_y(self):
        '''Returns the maximum y coordinate of the nodes of the mesh'''
        return max(self.nodes[:, 1])

    @property
    def coords_min_z(self):
        '''Returns the minimum z coordinate of the nodes of the mesh'''
        return min(self.nodes[:, 2])

    @property
    def coords_max_z(self):
        '''Returns the maximum z coordinate of the nodes of the mesh'''
        return max(self.nodes[:, 2])

    @property
    def region_names(self):
        '''Returns the names of the regions'''
        return list(self.region_dict.keys())

    @property
    def boundary_names(self):
        '''Returns the names of the boundaries'''
        return list(self.region_dict.keys())

    @property
    def material_names(self):
        '''Returns the names of the materials'''
        return list(self.material_dict.keys())

    def __repr__(self):
        import rich
        from rich.markdown import Markdown
        text = Markdown(f"""
         iGP mesh with {self.n_mesh_elements} elements and {self.n_mesh_nodes} nodes.
         Boundary names: {list(self.region_dict.keys())}
         Material names: {list(self.material_dict.keys())}
         """)
        rich.print(text)
        return ''

    def __str__(self):
        return self.__repr__()




def parallel_build_mesh_data(elements, nodes, shared_list, chunk_index, chunk_size, centroids):
    amount_read = 0.0
    for id_local, element in enumerate(elements):
        n_type = len(element)
        id_local_chunk = id_local + chunk_size * chunk_index
        if n_type == 4:  # This is a Tetrahedra object
            shared_list.append(TetrahedraElement(node_ids=element,
                                                 node_coords=nodes[element],
                                                 local_id=id_local_chunk,
                                                 centroid_coords=centroids[id_local_chunk] if config.general.constant_centroids else None,
                                                 # centroid_coords=self.centroids[id_local]
                                                 ))
        if n_type == 6:  # This is a Wedge object
            shared_list.append(WedgeElement(node_ids=element,
                                            node_coords=nodes[element],
                                            local_id=id_local_chunk,
                                            centroid_coords=centroids[id_local_chunk] if config.general.constant_centroids else None,
                                            # centroid_coords=self.centroids[id_local]
                                            ))
        if n_type == 8:  # This is a Hexahedra object
            shared_list.append(HexahedraElement(node_ids=element,
                                                node_coords=nodes[element],
                                                local_id=id_local_chunk,
                                                centroid_coords=centroids[id_local_chunk] if config.general.constant_centroids else None,
                                                # centroid_coords=self.centroids[id_local]
                                                ))
        if id_local / int(len(elements)) >= amount_read:
            amount = id_local / int(len(elements))
            logger.info(f"Process {chunk_index} completed amount: {amount * 100:3.0f} %")
            amount_read += 0.1













import logging

import numpy as np

from pydelling.config import config
from .BaseWriter import BaseWriter

logger = logging.getLogger(__name__)


class OpenFoamVariableWriter(BaseWriter):
    """This class creates an OpenFOAM variable data file"""
    def __init__(self, filename: str = None, header=None, outer=None, data=None, configuration_dict=None, *args, **kwargs):
        """
        A correct set of header/outer needs to be provided to the class.

        It can be automatically used from the config file by creating an instance called "open_foam_variable_writer" for the settings.

        Args:
            filename: Name of the output file
            header: Dictionary containing the header variables
            outer: Dictioanary containing the end variables
            data: Data to be set into the OpenFOAM file
        """
        if configuration_dict:
            self.header = configuration_dict.header
            self.outer = configuration_dict.outer
            self.filename = configuration_dict.filename
        else:
            self.header = header if header else config.open_foam_variable_writer.header
            self.outer = outer if outer else config.open_foam_variable_writer.outer
            self.filename = filename if filename else config.open_foam_variable_writer.filename
        self.data = data
        super().__init__(filename=self.filename, *args, **kwargs)

    def run(self, *args, **kwargs):
        """
        Writes the data into an OpenFOAM variable format
        """
        logger.info(f"Writing data to {self.filename}")
        with open(self.filename, "w") as self.output_file:
            self.write_header()
            self.write_data()
            self.write_outer()

    def write_header(self):
        """
        Writes the header of the file
        """
        self.output_file.write(f"""/*--------------------------------*- C++ -*----------------------------------*\\
  =========                 |
  \\\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\\    /   O peration     | Website:  https://openfoam.org
    \\\  /    A nd           | Version:  7
     \\\/     M anipulation  |
\*---------------------------------------------------------------------------*/
FoamFile
{{
    version     2.0;
    format      ascii;
    class       {self.header.field_type};
    location    "{self.header.location}";
    object      {self.header.object};
}}
// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

dimensions      {self.header.dimensions};
internalField   {self.header.data_type} {self.header.data_structure}\n""")

    def write_data(self):
        self.output_file.write(f"{len(self.data)}\n")
        self.output_file.write("(\n")
        for data_element in self.data:
            if type(data_element) == np.float64 or type(data_element) == np.float32:
                self.output_file.write(f"{data_element}\n")
            elif len(data_element) == 4:
                self.output_file.write(f"{data_element[3]}\n")
            else:
                logger.error("There is an error provided the data element vector")
                raise ValueError("There was an error providing the data element")

        self.output_file.write(")\n")
        self.output_file.write(";\n")

    def write_outer(self):
        self.output_file.write(f"""boundaryField\n{{\n""")
        for region in self.outer.boundary_fields:
            region_dict = self.outer.boundary_fields[region]
            self.output_file.write(f"""\t{region}\n\t{{\n\t\ttype\t{region_dict["type"]};\n\t}}\n""")
        self.output_file.write("}\n")
        self.output_file.write("\n")
        self.output_file.write("// ************************************************************************* //")



from .BaseAbstractMeshObject import BaseAbstractMeshObject
from .BaseElement import BaseElement
from .BaseFace import BaseFace
from .TriangleFace import TriangleFace
from .QuadrilateralFace import QuadrilateralFace
from .HexahedraElement import HexahedraElement
from .PyramidElement import PyramidElement
from .TetrahedraElement import TetrahedraElement
from .WedgeElement import WedgeElement



import unittest
from pydelling.preprocessing import DfnPreprocessor
from pydelling.utils import test_data_path
from pydelling.config import config


class DfnPreprocessorCase(unittest.TestCase):
    def test_dfn_sum(self):
        dfn_one = DfnPreprocessor()
        dfn_one.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.1, size=1.2
        )
        dfn_two = DfnPreprocessor()
        dfn_two.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.5, size=1.2
        )
        dfn_three = dfn_one + dfn_two
        self.assertEqual(len(dfn_three.dfn), 2)

    def test_dfn_save_load(self):
        dfn_one = DfnPreprocessor()
        dfn_one.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.1, size=1.2
        )
        dfn_one.to_json('test.json')
        dfn_one_loaded = DfnPreprocessor.from_json('test.json')
        self.assertEqual(len(dfn_one_loaded.dfn), 1)
        self.assertEqual(dfn_one_loaded.dfn[0].x_centroid, 0)
        self.assertEqual(dfn_one_loaded.dfn[0].aperture, 0.1)
        self.assertEqual(dfn_one_loaded.dfn[0].size, 1.2)

    def test_dfn_save_load_with_storativity(self):
        dfn_one = DfnPreprocessor()
        dfn_one.add_fracture(
            x=0, y=0, z=0, dip=0, dip_dir=0, aperture=0.1, size=1.2, rock_type=0, transmissivity_constant={0: 1E-6, 1: 1E-6},
        )
        dfn_one.to_json('test.json')
        dfn_one_loaded = DfnPreprocessor.from_json('test.json')
        self.assertEqual(len(dfn_one_loaded.dfn), 1)
        self.assertEqual(dfn_one_loaded.dfn[0].x_centroid, 0)
        self.assertEqual(dfn_one_loaded.dfn[0].aperture, 0.1)
        self.assertEqual(dfn_one_loaded.dfn[0].size, 1.2)

    def test_fault_save_load(self):
        dfn_one = DfnPreprocessor()
        dfn_one.add_fault(filename=test_data_path() / 'test_fault.stl', aperture=5.0)
        dfn_one.to_json('./test.json')
        dfn_one_loaded = DfnPreprocessor.from_json('test.json')
        self.assertEqual(len(dfn_one_loaded.faults), 1)
        self.assertEqual(dfn_one_loaded.faults[0].aperture, 5.0)

    def tearDown(self) -> None:
        import os
        try:
            os.remove('test.json')
        except FileNotFoundError:
            pass


if __name__ == '__main__':
    unittest.main()


import unittest
from pydelling.readers import RasterFileReader
from pydelling.utils import test_data_path


class TestSMeshMeshReader(unittest.TestCase):
    def setUp(self) -> None:
        self.top_surface = RasterFileReader(test_data_path() / 'top_surface.asc')
        self.bottom_surface = RasterFileReader(test_data_path() / 'bottom_surface.asc')

    def test_read_data(self):
        self.assertEqual(self.top_surface.nx, 250)
        self.assertEqual(self.top_surface.ny, 250)

    def test_operations(self):
        # constant sum
        sum = self.top_surface + 1
        self.assertEqual(self.top_surface.values[0, 2] + 1, sum.values[0, 2])
        # Constant subtraction
        diff = self.top_surface - 1
        self.assertEqual(self.top_surface.values[:, 2].mean() - 1, diff.values[:, 2].mean())
        # Constant multiplication
        prod = self.top_surface * 2
        self.assertEqual(self.top_surface.values[:, 2].mean() * 2, prod.values[:, 2].mean())
        # Constant division
        quot = self.top_surface / 2
        self.assertEqual(self.top_surface.values[:, 2].mean() / 2, quot.values[:, 2].mean())
        # Raster sum
        sum = self.top_surface + self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] + self.bottom_surface.values[0, 2], sum.values[0, 2])
        # Raster subtraction
        diff = self.top_surface - self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] - self.bottom_surface.values[0, 2], diff.values[0, 2])
        # Raster multiplication
        prod = self.top_surface * self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] * self.bottom_surface.values[0, 2], prod.values[0, 2])
        # Raster division
        quot = self.top_surface / self.bottom_surface
        self.assertEqual(self.top_surface.values[0, 2] / self.bottom_surface.values[0, 2], quot.values[0, 2])



if __name__ == '__main__':
    unittest.main()


from .BaseManager import BaseManager
from . import PflotranStudy
import subprocess
import docker
from docker.models.containers import Container
from io import BytesIO
from pathlib import Path
from pydelling.utils import create_results_folder
from pydelling.managers.PflotranPostprocessing import PflotranPostprocessing
import logging
import os

logger = logging.getLogger(__name__)
class PflotranManager(BaseManager):
    def _get_study_status(self, study_id: int):
        """This method returns the status of a study.
        """
        pass

    def _run_study(self,
                   study: PflotranStudy,
                   n_cores: int = 1):
        """This method runs a study.
        """
        # Change the working directory to the study folder
        os.chdir(study.output_folder)
        # test comment
        if n_cores == 1:
            # Run the study in serial
            subprocess.run(['pflotran', '-pflotranin', study.input_file_name])
        else:
            # Run the study in parallel
            subprocess.run(['$PETSC_DIR/$PETSC_ARCH/bin/mpirun', '-np', str(n_cores), 'pflotran', '-pflotranin', study.input_file_name])

    def _run_study_docker(self,
                          study: PflotranStudy,
                          docker_image: str,
                          n_cores: int = 1,
                          ):
        """This method runs a study using docker.
        """
        docker_client = docker.from_env()
            # Run the study in serial

        try:
            docker_client.api.start(self.manager_name)
            container = docker_client.containers.get(self.manager_name)
        except:
            docker_client.api.create_container(
                name=self.manager_name,
                image=docker_image,
                command='/bin/sh',
                detach=True,
                tty=True,
                volumes=[f'/home/pflotran/'],
                host_config=docker_client.api.create_host_config(
                    binds={
                        Path().cwd() / f'studies/{study.name}': {
                            'bind': f'/home/pflotran/{study.name}',
                            'mode': 'rw',
                        }
                    })
            )
            docker_client.api.start(container=self.manager_name)
            container = docker_client.containers.get(self.manager_name)

        # Compress study folder to a tar file
        bytes_io = BytesIO()
        import tarfile
        pw_tar = tarfile.TarFile(fileobj=bytes_io, mode='w')
        pw_tar.add(study.output_folder, arcname=study.output_folder.name)
        pw_tar.close()
        bytes_io.seek(0)
        # volume = docker_client.volumes.create(name=study.output_folder.name, driver='local', driver_opts={'type': 'none', 'device': study.output_folder, 'o': 'bind'})
        # Copy the tar file to the container
        container.exec_run(cmd="bash -c 'cd /home/pflotran'")
        container.put_archive('/home/pflotran', bytes_io)
        container.exec_run(cmd="bash -c 'ls'")
        if n_cores == 1:
            container.exec_run(cmd=f"sudo bash -c 'export PATH=$PATH:/home/pflotran/pflotran/src/pflotran &&"
                                   f" cd /home/pflotran/{study.output_folder.name} && "
                                   f"pflotran -pflotranin {study.input_file_name}'")
        else:
            result = container.exec_run(cmd=f"sudo bash -c 'export PATH=$PATH:/home/pflotran/pflotran/src/pflotran && "
                                   f"export PETSC_DIR=/home/pflotran/petsc && "
                                   f"export PETSC_ARCH=docker-petsc && "
                                   f"cd /home/pflotran/{study.output_folder.name} && "
                                   f"$PETSC_DIR/$PETSC_ARCH/bin/mpiexec -np {n_cores} pflotran -pflotranin {study.input_file_name}'")
        container.stop()
        container.remove()

    def merge_results(self, move=False, postprocess=True):
        """This method merges the results of all the studies.
        """
        self.run(dummy=True)
        import shutil
        # results_folder = create_results_folder(list(self.studies.values())[0].output_folder)
        self.merged_folder = create_results_folder(self.results_folder / 'merged_results')
        for study in self.studies.values():
            hdf5_files = list(study.output_folder.glob('*.h5'))
            for file in hdf5_files:
                if 'restart' not in file.name:
                    if not move:
                        shutil.copy(file, self.results_folder / 'merged_results')
                    else:
                        shutil.move(file, self.results_folder / 'merged_results')

        # Copy -domain.h5 file


        if postprocess:
            domain_folder = list(self.studies.values())[0].output_folder / 'input_files'
            domain_file = list(domain_folder.glob('*-domain.h5'))[0]
            shutil.copy(domain_file, self.results_folder / 'merged_results')
            logger.info('Postprocessing results')
            pflotran_postprocesser = PflotranPostprocessing()
            # Change the working directory to the results folder
            import os
            os.chdir(self.results_folder / 'merged_results')
            pflotran_postprocesser.run()
            # Return to the original working directory



















from pydelling.preprocessing.mesh_preprocessor import MeshPreprocessor


class UnstructuredMeshReader(MeshPreprocessor):
    def __init__(self, nodes, vertices):
        super().__init__()
        self.nodes = nodes
        self.vertices = vertices



import inspect
import subprocess
from abc import ABC
import streamlit as st
import extra_streamlit_components as stx
from .BaseStreamlitUtilityClass import BaseStreamlitUtilityClass
class WebAppRunner(ABC, BaseStreamlitUtilityClass):
    """This is the base class used to build other webapps."""
    def __init__(self):
        self.source_script_name = None
        if 'threading' not in inspect.stack()[-1][1]:
            self.source_script_name = inspect.stack()[-1][1]

    def construct(self):
        """This is the main method that builds the webapp."""
        pass

    def run(self):
        """This is the main method that runs the webapp."""
        current_executer = inspect.stack()
        if 'threading' not in current_executer[-1][1]:
            subprocess.run(["streamlit", "run", self.source_script_name])
        else:
            self.initialize()
            self.construct()

    def initialize(self):
        """This method initializes the webapp."""
        pass





import logging
from typing import List

import numpy as np

from ._AbstractGidObject import _AbstractGidObject

logger = logging.getLogger(__name__)

class Point(_AbstractGidObject):
    has_copy = False
    local_id: int = 1
    global_id: int = 1
    points: List = []
    def __init__(self, coords):
        self.local_id = Point.local_id
        self.id = None
        Point.local_id += 1
        if isinstance(coords, list):
            coords = np.array(coords)
            assert coords.shape[0] == 3, 'A 3D coordinate needs to be given'
        assert isinstance(coords, np.ndarray), 'Please provide a numpy array or a list'
        self.coords = coords

    def add(self):
        for point in Point.points:
            if np.array_equal(point.coords, self.coords):
                self.id = point.id
                self.has_copy = True
                return ''

        self.id = Point.global_id
        Point.global_id += 1
        Point.points.append(self)
        logger.info(f'Adding point {self.id} at {self.coords}')
        export_str = 'Mescape Geometry Create Point\n'
        export_str += f'{",".join(map(str, self.coords))}\n'
        return export_str

    @property
    def coords_comma(self):
        return f'{",".join(map(str, self.coords))}'

    def __repr__(self):
        if self.id:
            return f"Point {self.id} ({self.coords})"
        else:
            return f"Point (local) {self.local_id} ({self.coords})"






from __future__ import annotations

import logging
from typing import *

import meshio as msh
import numpy as np
from scipy.spatial import KDTree
from tqdm import tqdm

import pydelling.preprocessing.mesh_preprocessor.geometry as geometry
from pydelling.preprocessing.dfn_preprocessor.Fracture import Fracture
from pydelling.preprocessing.mesh_preprocessor.geometry import BaseElement
from pydelling.utils.geometry_utils import compute_polygon_area

logger = logging.getLogger(__name__)


class MeshPreprocessor(object):
    """Contains the logic to preprocess and work with a generic unstructured mesh"""
    elements: List[geometry.BaseElement]
    coords: List[np.ndarray]
    centroids: List[np.ndarray]
    meshio_mesh: msh.Mesh = None
    kd_tree: KDTree = None
    point_data = {}
    cell_data = {}
    _coords = None
    _centroids = None
    is_intersected = False
    is_streamlit = False
    aux_nodes = {}
    has_kd_tree: bool = False

    def __init__(self, *args, **kwargs):
        self.unordered_nodes = {}
        self.elements = []
        if 'st_file' in kwargs:
            self.is_streamlit = True

        self.find_intersection_stats = {
            'total_intersections': 0,
            'intersection_points':
                {
                }
        }

    def add_element(self, element: geometry.BaseElement):
        self.elements.append(element)

    def add_tetrahedra(self, node_ids: List[int] or np.ndarray, node_coords: List[np.ndarray]):
        """Adds a tetrahedron to the mesh"""
        self.elements.append(geometry.TetrahedraElement(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_hexahedra(self, node_ids: List[int] or np.ndarray, node_coords: List[np.ndarray]):
        """Adds a hexahedron to the mesh"""
        self.elements.append(geometry.HexahedraElement(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_wedge(self, node_ids: List[int] or np.ndarray, node_coords: List[np.ndarray]):
        """Adds a wedge to the mesh"""
        self.elements.append(geometry.WedgeElement(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_pyramid(self, node_ids: List[int] or np.ndarray, node_coords: List[np.ndarray]):
        """Adds a pyramid to the mesh"""
        self.elements.append(geometry.PyramidElement(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_triangular_prism(self, node_ids: List[int] or np.ndarray, node_coords: List[np.ndarray]):
        """Adds a triangular prism to the mesh"""
        self.elements.append(geometry.WedgeElement(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    @property
    def coords(self) -> np.ndarray:
        """Orders the coords in the mesh"""
        if self._coords is None:
            aux_nodes = np.ndarray(shape=(self.n_nodes, 3))
            for idx, node in self.unordered_nodes.items():
                aux_nodes[idx] = node
            self._coords = aux_nodes
        return self._coords

    def add_quadrilateral(self, node_ids: List[int], node_coords: List[np.ndarray]):
        self.elements.append(geometry.QuadrilateralFace(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_triangle(self, node_ids: List[int], node_coords: List[np.ndarray]):
        self.elements.append(geometry.TriangleFace(node_ids=node_ids, node_coords=node_coords))
        for idx, node in enumerate(node_coords):
            self.unordered_nodes[node_ids[idx]] = node

    def add_node(self, node: np.ndarray):
        '''Explicitly adds a node (deprecated)'''
        self.coords.append(node)

    @property
    def n_nodes(self):
        """Returns the number of node_ids in the mesh"""
        return len(self.unordered_nodes)

    @property
    def n_elements(self):
        """Returns the number of elements in the mesh"""
        return len(self.elements)

    def convert_mesh_to_meshio(self):
        """Converts the mesh into a meshio mesh
        Returns:
            meshio mesh
        """

        elements_in_meshio = self._create_meshio_dict(self.elements)
        self.meshio_mesh = msh.Mesh(
            points=self.coords,
            cells=elements_in_meshio,
            cell_data=self.cell_data,
            point_data=self.point_data
        )

    def to_vtk(self, filename='mesh.vtk'):
        """Converts the mesh into vtk using meshio
        Args:
            filename: name of the output file
        """
        logger.info(f'Converting mesh to vtk and exporting to {filename}')
        self.convert_mesh_to_meshio()
        self.meshio_mesh.write(filename)

    def subset_to_vtk(self, elements: List[geometry.BaseAbstractMeshObject], filename='subset.vtk'):
        """Converts a subset of the mesh into a vtk file
        Args:
            elements: list of element indices
            filename: name of the output file
        """
        assert len(elements) > 0, 'No elements to export'
        subset_mesh = self._convert_subset_to_meshio(elements)
        subset_mesh.write(filename)

    def _convert_subset_to_meshio(self, elements: List[geometry.BaseAbstractMeshObject]) -> msh.Mesh:
        """Converts a subset of the mesh into a meshio mesh
        Args:
            elements: list of element indices
        Returns:
            meshio mesh
        """
        elements_in_meshio = self._create_meshio_dict(elements)
        return msh.Mesh(
            points=self.coords,
            cells=elements_in_meshio,
        )

    def _create_meshio_dict(self, elements: List[geometry.BaseAbstractMeshObject]) -> Dict[str, List[List[int]]]:
        elements_in_meshio = {}
        for element in elements:
            if element.type == 'tetrahedra':
                if not 'tetra' in elements_in_meshio.keys():
                    elements_in_meshio['tetra'] = []
                elements_in_meshio['tetra'].append(element.nodes.tolist())
            elif element.type == 'hexahedra':
                if not 'hexahedron' in elements_in_meshio.keys():
                    elements_in_meshio['hexahedron'] = []
                elements_in_meshio['hexahedron'].append(element.nodes.tolist())
            elif element.type == 'wedge':
                if not 'wedge' in elements_in_meshio.keys():
                    elements_in_meshio['wedge'] = []
                elements_in_meshio['wedge'].append(element.nodes.tolist())
            elif element.type == 'triangle':
                if not 'triangle' in elements_in_meshio.keys():
                    elements_in_meshio['triangle'] = []
                elements_in_meshio['triangle'].append(element.nodes.tolist())
            elif element.type == 'quadrilateral':
                if not 'quad' in elements_in_meshio.keys():
                    elements_in_meshio['quad'] = []
                elements_in_meshio['quad'].append(element.nodes.tolist())
            elif element.type == 'pyramid':
                if not 'pyramid' in elements_in_meshio.keys():
                    elements_in_meshio['pyramid'] = []
                elements_in_meshio['pyramid'].append(element.nodes.tolist())

        return elements_in_meshio

    def nodes_to_csv(self, filename='node_ids.csv'):
        """Exports the node_ids to CSV"""
        node_array = np.array(self.coords)
        np.savetxt(filename, node_array, delimiter=',')

    @property
    def centroids(self):
        """Returns the centroids of the mesh"""
        if self._centroids is None:
            centroids = []
            for element in self.elements:
                centroids.append(element.centroid)
            self._centroids = np.array(centroids)
        return self._centroids

    def create_kd_tree(self, kd_tree_config=None):
        """
        Create a KD-tree structure for the mesh.
        Args:
            kd_tree_config: A dictionary with the kd-tree configuration.
        """
        if kd_tree_config is None:
            kd_tree_config = {}
        self.kd_tree = KDTree(self.centroids, **kd_tree_config)

    def get_k_nearest_mesh_elements(self, point, k=15, distance_upper_bound=None):
        """
        Get the nearest mesh elements to a point.
        Args:
            point: A point in 3D space.
            k: The number of nearest elements to return.
        Returns:
            A list of the nearest mesh elements.
        """
        if not hasattr(self, 'kd_tree'):
            self.create_kd_tree()
        if distance_upper_bound:
            ids = self.kd_tree.query(point, k=k, distance_upper_bound=distance_upper_bound)[1]
        else:
            ids = self.kd_tree.query(point, k=k)[1]

        assert len(ids) != 0, "No elements found"
        return [self.elements[i] for i in ids]

    def get_closest_mesh_elements(self, point, distance=None):
        """
        Get the nearest mesh elements to a point inside a distance.
        Args:
            point: A point in 3D space.
            distance: The radius of the sphere.
        Returns:
            A list of the nearest mesh elements.
        """
        if self.kd_tree is None:
            self.create_kd_tree()

        ids = self.kd_tree.query_ball_point(point, distance)
        # assert len(ids) != 0, "No elements found"
        return [self.elements[i] for i in ids]

    def get_closest_n_mesh_elements(self,
                                    point,
                                    n=1
                                    ):
        """
        Get the nearest mesh elements to a point inside a distance.
        Args:
            point: A point in 3D space.
            distance: The radius of the sphere.
        Returns:
            A list of the nearest mesh elements.
        """
        if self.kd_tree is None:
            self.create_kd_tree()

        ids = self.kd_tree.query(point, k=n)[1]
        return [self.elements[i] for i in ids]

    def clear(self):
        """Clears the mesh"""
        self.unordered_nodes = {}
        self.elements = []

    @staticmethod
    def _intersect_fracture_with_element(element, fracture):
        """
        Returns the intersection of a fracture with an element.
        Args:
            element: The element to intersect with.
            fracture: The fracture to intersect with.
        Returns:
            The intersection of the fracture and the element.
        """
        return element.intersect(fracture)

    def _is_fracture_intersected(self, fracture: Fracture, element: geometry.BaseElement):
        """
        Checks if a fracture is intersected by an element.
        Args:
            fracture: The fracture to intersect with.
            element: The element to intersect with.
        Returns:
            True if the fracture is intersected by the element, False otherwise.
        """
        signs = []
        bounding_box: List = fracture.get_bounding_box()
        for coord in element.coords:
            # Check if coord in bounding box
            if bounding_box[0] < coord[0] < bounding_box[1] and bounding_box[2] < coord[1] < bounding_box[3] and \
                    bounding_box[4] < coord[2] < bounding_box[5]:

                distance_to_fracture = fracture.distance_to_point(coord)
                signs.append(np.sign(distance_to_fracture))
            else:
                return False

        if not np.all(np.array(signs) == signs[0]):
            return True


    def find_the_intersection_between_fracture_and_mesh(self, fracture: Fracture):
        """Finds the intersection between a fracture and the mesh"""
        intersections = []
        for element in self.elements:
            if self._is_fracture_intersected(fracture, element):
                intersections.append(element)
        self.subset_to_vtk(intersections, filename='intersections.vtk')

    def find_intersection_points_between_fracture_and_mesh(self, fracture: Fracture, export_stats=False):
        """Finds the intersection points between a fracture and the mesh"""

        intersection_points = []
        kd_tree_filtered_elements = self.get_closest_mesh_elements(fracture.centroid, distance=fracture.size)
        counter = 0
        for element in kd_tree_filtered_elements:
            element: geometry.BaseElement
            counter += 1
            intersection_points = element.intersect_with_fracture(fracture)

            if len(intersection_points) >= 3:
                intersection_area = compute_polygon_area(intersection_points)
                fracture.intersection_dictionary[element.local_id] = intersection_area
                element.associated_fractures[fracture.local_id] = {
                    'area': intersection_area,
                    'volume': intersection_area * fracture.aperture,
                    'fracture': fracture,
                }
            n_intersections = len(intersection_points)
            if not n_intersections in self.find_intersection_stats['intersection_points'].keys():
                self.find_intersection_stats['intersection_points'][n_intersections] = 0
            self.find_intersection_stats['intersection_points'][n_intersections] += 1
            self.find_intersection_stats['total_intersections'] += 1

        self.is_intersected = True

        return intersection_points


    def export_intersection_stats(self, filename='intersection_stats.txt'):
        # Export the run_stats dictionary to file
        assert self.is_intersected, 'The mesh has not been intersected yet.'
        import json
        with open('run_stats.json', 'w') as fp:
            json.dump(self.find_intersection_stats, fp)

    @staticmethod
    def intersect_edge_plane(edge: np.ndarray,
                             edge_point: np.ndarray,
                             plane: Fracture,
                             ) -> np.ndarray or None:
        """This method instersects a given edge with a plane.
        Args:
            edge: The edge to intersect.
            plane: The plane to intersect with.
        Returns:
            The intersection point of the edge and the plane.
        """

        edge_dot = np.dot(edge, plane.unit_normal_vector)
        if edge_dot == 0:
            return None
        else:
            t = -np.dot((edge_point - plane.centroid), plane.unit_normal_vector) / edge_dot
            if t < 1.0 and t > 0.0:
                point = edge_point + t * edge
                if plane.point_inside_bounding_box(point):
                    return point
            else:
                return None

    @property
    def min_x(self):
        return self.coords[:, 0].min()

    @property
    def max_x(self):
        return self.coords[:, 0].max()

    @property
    def min_y(self):
        return self.coords[:, 1].min()

    @property
    def max_y(self):
        return self.coords[:, 1].max()

    @property
    def min_z(self):
        return self.coords[:, 2].min()

    @property
    def max_z(self):
        return self.coords[:, 2].max()


    def save(self, filename):
        """Save the mesh to a file."""
        import pickle
        logger.info(f'Saving mesh to {filename}')
        with open(filename, 'wb') as f:
            save_dictionary = {
                'elements': self.elements,
                'coords': self.coords,
                'kd_tree': self.kd_tree,
                'has_kd_tree': self.has_kd_tree,
            }
            pickle.dump(save_dictionary, f)

    def load(self, filename):
        """Load the mesh from a file."""
        logger.info(f'Loading mesh from {filename}')
        import pickle
        with open(filename, 'rb') as f:
            save_dictionary = pickle.load(f)
            self.elements = save_dictionary['elements']
            self._coords = save_dictionary['coords']
            self.kd_tree = save_dictionary['kd_tree']
            self.has_kd_tree = save_dictionary['has_kd_tree']

    def get_json(self):
        """Export the mesh to a json file."""
        logger.info('Exporting mesh to json')
        save_dictionary = {}
        _elements = [element.get_json() for element in self.elements]
        save_dictionary['elements'] = _elements
        save_dictionary['coords'] = self.coords.tolist()
        save_dictionary['has_kd_tree'] = self.has_kd_tree
        return save_dictionary

    def to_json(self, filename='mesh.json'):
        """Export the mesh to a json file."""
        import json
        with open(filename, 'w') as f:
            json.dump(self.get_json(), f)

    @classmethod
    def from_json(self, filename='mesh.json'):
        """Load the mesh from a json file."""
        logger.info(f'Loading mesh from {filename}')
        BaseElement.local_id = 0
        import json
        with open(filename, 'r') as f:
            save_dictionary = json.load(f)
            mesh = MeshPreprocessor()
            mesh._coords = np.array(save_dictionary['coords'])
            mesh.has_kd_tree = save_dictionary['has_kd_tree']
            MeshPreprocessor.load_elements(mesh, save_dictionary['elements'])
            if mesh.has_kd_tree:
                mesh.create_kd_tree()
            return mesh

    @classmethod
    def from_dict(cls, dict: dict):
        """Load the mesh from a json file."""
        BaseElement.local_id = 0
        mesh = MeshPreprocessor()
        mesh._coords = np.array(dict['coords'])
        mesh.has_kd_tree = dict['has_kd_tree']
        MeshPreprocessor.load_elements(mesh, dict['elements'])
        if mesh.has_kd_tree:
            mesh.create_kd_tree()
        return mesh

    @staticmethod
    def load_elements(mesh: MeshPreprocessor, element_dict):
        """Load the elements from a dictionary."""
        elements = []
        for local_id, element in tqdm(enumerate(element_dict), desc='Loading elements'):
            if element['type'] == 'tetrahedra':
                mesh.add_tetrahedra(node_ids=element['nodes'],
                                    node_coords=mesh._coords[element['nodes']])
            elif element['type'] == 'hexahedra':
                mesh.add_hexahedra(node_ids=element['nodes'],
                                   node_coords=mesh._coords[element['nodes']])
            elif element['type'] == 'wedge':
                mesh.add_wedge(node_ids=element['nodes'],
                               node_coords=mesh._coords[element['nodes']])
            elif element['type'] == 'pyramid':
                mesh.add_pyramid(node_ids=element['nodes'],
                                 node_coords=mesh._coords[element['nodes']])
            associated_fractures_dict = element_dict[local_id]['associated_fractures']
            temp_associated_fractures = {}
            for key in associated_fractures_dict:
                cur_fracture = associated_fractures_dict[key]
                temp_fracture = {key: value for key, value in cur_fracture.items() if key != 'fracture'}
                temp_fracture['fracture'] = cur_fracture['fracture']
                temp_associated_fractures[key] = temp_fracture
            mesh.elements[local_id].associated_fractures = temp_associated_fractures



    def refactor_array_by_element_type(self, array: np.ndarray or list) -> list:
        """Refactors a given array based on the element type"""
        if isinstance(array, np.ndarray):
            array = array.tolist()
        final_array = []
        temp_dict = {
            'wedge': [],
            'pyramid': [],
            'tetrahedra': [],
            'hexahedra': [],
        }
        for element in self.elements:
            temp_dict[element.type].append(array[element.local_id])
        for key in temp_dict:
            if len(temp_dict[key]) > 0:
                final_array.append(temp_dict[key])
        return final_array

    def __repr__(self):
        return f'Mesh with {len(self.elements)} elements and {len(self.coords)} nodes.'

    def add_cell_data(self, name, data):
        self.cell_data[name] = self.refactor_array_by_element_type(data)

    def add_point_data(self, name, data):
        self.point_data[name] = data




from .BaseStudy import BaseStudy
from abc import ABC, abstractmethod
from typing import Dict, List, Union
import logging
from alive_progress import alive_bar
from tqdm import tqdm
from pydelling.utils import create_results_folder
from docker import DockerClient


logger = logging.getLogger(__name__)


class BaseManager(ABC):
    """This class is the base class for all simulation managers. A simulation manager should be able to:
    - Manage study [BaseStudy] objects
    - Run simulations on a given software
    - Read the output status of the simulation
    """
    def __init__(self, name: str = None):
        self.results_folder = None
        self.studies: Dict[str, BaseStudy] = {}
        self.manager_name = name if name is not None else self.__class__.__name__
        self.is_dummy = False

    def add_study(self, study: BaseStudy):
        """This method adds a study to the manager.
        """
        assert isinstance(study, BaseStudy), f"Study must be a object from a class inherited from BaseStudy, not {type(study)}"
        self.studies[study.name] = study

    def run(self,
            studies_folder: str = './studies',
            n_cores: int = 1,
            docker_image: str = None,
            dummy: bool = False,
            ):
        """This method runs all the studies.
        """
        self.is_dummy = dummy
        # Initialize callbacks
        for study in self.studies.values():
            study.initialize_callbacks(self)

        self.results_folder = create_results_folder(studies_folder)
        # self.generate_run_files(studies_folder=studies_folder)
        for study in tqdm(self.studies.values(), desc="Running studies", colour="white"):
            self.run_study(study, docker_image=docker_image, n_cores=n_cores, dummy=dummy)

    def generate_run_files(self, studies_folder: str = './studies'):
        """This method generates the run files for all the studies.
        """
        for study in self.studies.values():
            study: BaseStudy
            study.to_file(self.results_folder / study.name)

    @abstractmethod
    def _run_study(self, study: BaseStudy, n_cores: int = 1):
        """This method runs a study.
        """
        logger.info(f"Running study {study.name}")
        pass

    @abstractmethod
    def _run_study_docker(self,
                          study: BaseStudy,
                          docker_image: str,
                          n_cores: int = 1,
                          ):
        """This method runs a study using docker.
        """
        logger.info(f"Running study {study.name} using docker image {docker_image}")

    def run_study(self,
                  study: BaseStudy,
                  n_cores: int = 1,
                  docker_image: str = None,
                  dummy: bool = False,
                  ):
        """This method runs a study.
        """
        logger.info(f"Running study {study.name}")
        # Create the study files
        study.output_folder = self.results_folder / study.name
        study.pre_run()
        for callback in study.callbacks:
            if callback.kind == 'pre':
                callback.run()

        study.to_file(self.results_folder / study.name)
        # Run the study
        if docker_image is None:
            if dummy:
                logger.info("Dummy run, not running the study")
            else:
                self._run_study(study, n_cores=n_cores)
        else:
            self._run_study_docker(study, docker_image, n_cores=n_cores)

        for callback in study.callbacks:
            if callback.kind == 'post':
                callback.run()
        study.post_run()


    @property
    def n_studies(self):
        """This method returns the number of studies.
        """
        return len(self.studies)

    def merge_studies(self):
        """This methods merges the studies generated by the manager.
        """
        pass








from unittest import TestCase
from pydelling.managers import PflotranStudy
from pydelling.utils.configuration_utils import test_data_path


class TestPflotranCase(TestCase):
    def setUp(self) -> None:
        self.pflotran_study = PflotranStudy(str(test_data_path() / 'test_manager.in'))

    def test_get_regions(self):
        regions = self.pflotran_study.get_regions()
        self.assertEqual(regions, ['all', 'inlet', 'sides', 'top_below', 'top_above', 'bottom', 'initial', 'Sand'])

    def test_get_time(self):
        time = self.pflotran_study.get_simulation_time(time_unit='d')
        self.assertEqual(time, 365.0)
        new_time = 2
        new_time_unit = 'd'
        self.pflotran_study.replace_simulation_time(new_time=new_time, time_unit=new_time_unit)
        time = self.pflotran_study.get_simulation_time(time_unit=new_time_unit)
        self.assertEqual(time, new_time)

    def test_get_region_file(self):
        region = 'all'
        file = self.pflotran_study.get_region_file(region)
        self.assertEqual(file, None)
        region = 'top_below'
        file = self.pflotran_study.get_region_file(region)
        self.assertEqual(file, './input_files/top_below.ex')

    def test_replace_region_file(self):
        region = 'top_below'
        new_file = './input_files/test_replace.ex'
        self.pflotran_study.replace_region_file(region, new_file)
        file = self.pflotran_study.get_region_file(region)
        self.assertEqual(file, new_file)

    def test_checkpoint(self):
        times = [1, 2, 3]
        time_unit = 'd'
        self.pflotran_study.add_checkpoint(times=times, time_unit=time_unit)
        checkpoints = self.pflotran_study.get_checkpoint()
        self.assertEqual(checkpoints, 'TIMES d 1.0 2.0 3.0')
        manager_no_checkpoint = PflotranStudy(str(test_data_path() / 'test_pflotran_manager_nocheckpoint.in'))
        self.assertEqual(manager_no_checkpoint.get_checkpoint(), None)
        manager_no_checkpoint.add_checkpoint(times=times, time_unit=time_unit)
        self.assertEqual(manager_no_checkpoint.get_checkpoint(), 'TIMES d 1.0 2.0 3.0')
        manager_no_checkpoint.to_file()

    def test_datasets(self):
        datasets = ['dirichletpressure', 'topflow']
        self.assertEqual(self.pflotran_study.get_datasets(), datasets)
        self.pflotran_study.add_dataset(name='dirichletpressure', filename='pepe.dat', hdf5_dataset_name='pepe')
        self.pflotran_study.add_dataset(name='pepe', filename='pepe.dat', hdf5_dataset_name='pepe')
        new_datasets = ['dirichletpressure', 'topflow', 'pepe']
        self.assertListEqual(sorted(self.pflotran_study.get_datasets()), sorted(new_datasets))

















"""
Interpolates a given set of points into a PFLOTRAN mesh
"""
import logging

import numpy as np
from scipy.interpolate import griddata

from pydelling.utils.decorators import set_run
from .BaseInterpolator import BaseInterpolator

logger = logging.getLogger(__name__)


class SparseDataInterpolator(BaseInterpolator):
    divide_over_direction = None
    @set_run
    def run(self, method='nearest', divide_over_direction=None, **kwargs):
        """
        Interpolates the data using the given method.
        Args:
            method: interpolation method
            divide_over_direction: Divides the data into smaller chunks if the data is too large.
            **kwargs:

        Returns:
            interpolated data
        """
        if not divide_over_direction:
            logger.info(f"Interpolating data based on {self.info}")
            self.interpolated_data = griddata(self.data[:, 0:-1], self.data[:, -1], self.mesh, method=method, **kwargs)
            return self.get_data()
        else:
            logger.info(f"Dividing data into smaller chunks")
            # Divide the data depending on the given direction
            # For now, only divide in the x direction using the mean value
            mesh_x_plus: np.ndarray = self.mesh[self.mesh[:, 0] >= self.mesh[:, 0].mean()]
            mesh_x_minus: np.ndarray = self.mesh[self.mesh[:, 0] < self.mesh[:, 0].mean()]
            data_x_plus: np.ndarray = self.data[self.data[:, 0] >= self.data[:, 0].mean()]
            data_x_minus: np.ndarray = self.data[self.data[:, 0] < self.data[:, 0].mean()]
            # Interpolate the data
            interpolate_plus = griddata(data_x_plus[:, 0:-1], data_x_plus[:, -1], mesh_x_plus, method=method, **kwargs)
            interpolate_minus = griddata(data_x_minus[:, 0:-1], data_x_minus[:, -1], mesh_x_minus, method=method, **kwargs)
            # Combine the data
            self.interpolated_data = np.concatenate((interpolate_plus, interpolate_minus), axis=0)
            return self.get_data()




    def get_data(self):
        temp_array = np.reshape(self.interpolated_data, (self.interpolated_data.shape[0], 1))
        return np.concatenate((self.mesh, temp_array), axis=1)

    def change_min_value(self, min_value=None):
        logger.info(f"Equaling values <{min_value} to {min_value}")
        self.interpolated_data[self.interpolated_data < min_value] = min_value
        return self.interpolated_data


    def plot_regular_mesh(self):
        """Generates a plot of the interpolated data on a regular mesh"""
        assert self.is_run, "The interpolator has not been run"
        assert self.has_regular_mesh, "The interpolator has not been run with a regular mesh"
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        plot_data = self.get_data()


    def generate_pointwise_data(self):
        """Generates a pointwise data file"""
        assert self.is_run, "The interpolator has not been run"
        assert self.has_regular_mesh, "The interpolator has not been run with a regular mesh"
        mesh_data = self.mesh.copy()
        mesh_data = np.hstack((mesh_data, np.reshape(self.interpolated_data, (-1, 1))))
        return mesh_data

    def export_pointwise_data(self, output_file='pointwise_data.csv'):
        """Exports the interpolated data to a csv data file"""
        assert self.is_run, "The interpolator has not been run"
        assert self.has_regular_mesh, "The interpolator has not been run with a regular mesh"
        mesh_data = self.generate_pointwise_data()
        np.savetxt(output_file, mesh_data, delimiter=',', header='x,y,value')